"""Testing BaseModel class
"""
import datetime
import os
import pathlib
import sys
import uuid
from collections import OrderedDict
from copy import deepcopy

import pytest

from releso.base_model import BaseModel, add_save_location_if_elem_is_o_dict

base_path = str(pathlib.Path(__file__).parent)

dir_save_location_local = str(
    (
        pathlib.Path(__file__).parent / "test_save_location_please_delete"
    ).resolve()
)


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10")
@pytest.mark.parametrize(
    "test_input, changes",
    [
        (OrderedDict(), True),
        (
            OrderedDict([("save_location", "something")]),
            False,
        ),
        (
            OrderedDict(
                [("test", OrderedDict()), ("save_location", "something")]
            ),
            False,
        ),
        (["test", "asdas", 12, 12.0], False),
        ({"test": "asdas", "test2": 12, "test3": 12.0}, True),
        (
            {"test": "asdas", "test2": 12, "test3": 12.0, "save_location": 12},
            False,
        ),
        ([OrderedDict(), OrderedDict()], True),
        ([OrderedDict(), "testing"], True),
    ],
)
def test_base_model_add_save_location_if_elem_is_o_dict(test_input, changes):
    input_copy = deepcopy(test_input)
    save_location_str = "asdasdedascsadwasdvril324"
    add_save_location_if_elem_is_o_dict(test_input, save_location_str)
    assert (test_input == input_copy) is not changes
    if changes:
        if isinstance(input_copy, list):
            for idx in range(len(input_copy)):
                if isinstance(test_input[idx], OrderedDict | dict):
                    assert (
                        test_input[idx]["save_location"] == save_location_str
                    )
        else:
            assert test_input["save_location"] == save_location_str
    else:
        if isinstance(input_copy, OrderedDict | dict):
            assert test_input["save_location"] != save_location_str


@pytest.mark.parametrize(
    "test_input",
    [
        (dir_save_location_local),
        (f"please_delete_{uuid.uuid4()}"),
        ("please_delete_{}"),  # this test might fail if the time is exactly
    ],
)
def test_base_model_convert_to_pathlib_add_datetime(test_input):
    simple_path = None
    if "{}" not in test_input:
        if (simple_path := pathlib.Path(test_input)).is_dir():
            os.rmdir(simple_path)
    else:
        simple_path = pathlib.Path(
            test_input.format(
                datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            )
        ).resolve()
    result = BaseModel.convert_to_pathlib_add_datetime(test_input)
    if simple_path:
        assert result == simple_path.resolve()
    assert isinstance(result, pathlib.Path)
    assert result.is_dir()
    os.rmdir(result)


@pytest.mark.parametrize(
    "logger_name",
    [
        (None),
        ("this_is_a_logger_name"),
    ],
)
def test_base_model_initiate(logger_name, dir_save_location):
    model = BaseModel(save_location=dir_save_location, logger_name=logger_name)

    assert model.save_location == pathlib.Path(dir_save_location)
    assert model.logger_name == logger_name


@pytest.mark.parametrize("original_logger_name", [None, "original_name"])
def test_base_model_set_logger_name_recursively(
    original_logger_name, dir_save_location
):
    model = BaseModel(
        save_location=dir_save_location, logger_name=original_logger_name
    )

    model.__dict__["test"] = deepcopy(model)
    model.__dict__["test2"] = [
        deepcopy(model),
        123,
        [deepcopy(model), 123, 123.0, "test"],
    ]

    model.set_logger_name_recursively("new_logger_name")

    assert model.logger_name == "new_logger_name"


def test_base_model_multiprocessing_logger(dir_save_location):
    model = BaseModel(
        save_location=dir_save_location_local,
        logger_name="original_logger_name_multiprocessing",
    )

    logger = model.get_logger()

    assert logger.name == "multiprocessing"



import numpy as np
import pytest
from gymnasium import spaces

from releso.observation import (
    ObservationDefinition,
    ObservationDefinitionMulti,
)


@pytest.mark.parametrize(
    "name, value_min, value_max, expected",
    [
        ("test", -1, 1, spaces.Box(-1, 1, shape=(1,), dtype=np.float32)),
        ("test", 0, 12, spaces.Box(0, 12, shape=(1,), dtype=np.float32)),
    ],
)
def test_observation_definition(
    name, value_min, value_max, expected, dir_save_location
):
    observation = ObservationDefinition(
        name=name,
        value_min=value_min,
        value_max=value_max,
        save_location=dir_save_location,
    )
    assert observation.get_observation_definition()[1] == expected
    assert np.allclose(
        observation.get_default_observation(), np.array([value_min])
    )


@pytest.mark.parametrize(
    "name, value_min, value_max, observation_shape, value_type, expected",
    [
        (
            "test",
            -1,
            1,
            [1],
            "float",
            spaces.Box(-1, 1, shape=(1,), dtype=np.float32),
        ),
        (
            "test",
            0,
            12,
            [1],
            "float",
            spaces.Box(0, 12, shape=(1,), dtype=np.float32),
        ),
        (
            "test",
            0,
            12,
            [12, 12],
            "float",
            spaces.Box(0, 12, shape=(12, 12), dtype=np.float32),
        ),
        (
            "test",
            0,
            255,
            [224, 224],
            "CNN",
            spaces.Box(0, 255, shape=(224, 224), dtype=np.uint8),
        ),
    ],
)
def test_multi_observation_definition(
    name,
    value_min,
    value_max,
    observation_shape,
    value_type,
    expected,
    dir_save_location,
):
    observation = ObservationDefinitionMulti(
        name=name,
        value_min=value_min,
        value_max=value_max,
        observation_shape=observation_shape,
        value_type=value_type,
        save_location=dir_save_location,
    )
    assert observation.get_observation_definition()[1] == expected

    if value_type == "float":
        assert np.allclose(
            np.ones(observation_shape) * value_min,
            observation.get_default_observation(),
        )
    elif value_type == "CNN":
        assert np.allclose(
            np.zeros(observation_shape, np.uint8),
            observation.get_default_observation(),
        )
    else:
        RuntimeError("Unknown value type")



"""SPOR object script helper functions.

This files holds functions which can be used in custom command line spor object
scripts.
Tried to keep this as package independent as possible, to make it easier to
copy these functions into your own scripts. If you need to have other python
versions/special environment to run your scripts in. The pydantic dependency
can be exchanged for a string.
"""
import argparse
import json
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic.types import UUID4


def spor_com_parse_arguments(
    own_arguments: Optional[List[str]] = None,
) -> argparse.Namespace:
    """Parses the spor com arguments.

    The result is returned in a namespace object for easy access.

    Returns:
        argparse.Namespace: [description]
    """
    parser = argparse.ArgumentParser(
        description="Reinforcement Learning based Shape Optimization"
        " Toolbox. This is the basic script that can load a json file and run"
        " the resulting optimization problem."
    )
    parser.add_argument(
        "-i",
        "--initialize",
        "--reset",
        dest="reset",
        action="store_true",
        help="Should the script perform the step necessary for a reset.",
    )
    parser.add_argument(
        "-r",
        "--run",
        "--run_id",
        dest="run_id",
        action="store",
        type=UUID4,  # can also be a string. If pydantic is not available.
        required=True,
        help="Id of the SPOR_Communication interface for this "
        "specific spor object.",
    )
    parser.add_argument(
        "-v",
        "--validation_value",
        dest="validation_value",
        action="store",
        type=float,
        required=False,
        help="If during validation the current validation value is passed to "
        "the script by the variable. If not present not currently in "
        "validation.",
    )
    parser.add_argument(
        "-j",
        "--additional_values",
        "--json_object",
        dest="json_object",
        action="store",
        type=str,
        required=False,
        help="Currently available step information including observations, "
        "done, reward, info. Will automatically be parsed into a dict. Only "
        "present if spor step is configured to sent it.",
    )
    parser.add_argument(
        "-l",
        "--base_save_location",
        dest="base_save_location",
        action="store",
        type=str,
        required=True,
        help="Path pointing to the save location of all permanent records of "
        "this trainings run. Please save the logs here.",
    )
    parser.add_argument(
        "-e",
        "--environment_id",
        dest="environment_id",
        action="store",
        type=str,
        required=True,
        help="ID of the environment this call is "
        "coming from. NOT unique to this spor object.",
    )
    if own_arguments is not None:
        args = parser.parse_args(own_arguments)
    else:
        args = parser.parse_args()  # pragma: no cover
    if args.json_object:
        args.json_object = spor_com_additional_information(args.json_object)
    return args


def spor_com_additional_information(j_str: str) -> Dict[str, Any]:
    """Converts the given string (should hold a json string) into a dictionary.

    This function is currently only used for the spor_com_parse_arguments
    function.

    Args:
        j_str (str): string holding a json definition

    Returns:
        Dict[str, Any]: Dict created from the json definition
    """
    a = json.loads(j_str[1:-1])
    return a


def load_json(f_n: Union[PathLike, str]) -> Dict[str, Any]:
    """Loads data from a given file as a json object.

    Will create a dummy file if it does not exist.

    Args:
        f_n (Union[PathLike, str]): Path to the file where the json encoded
        data is stored.

    Returns:
        Dict[str, Any]: Data loaded from the given file.
    """
    path = Path(f_n)
    if not (path.exists() and path.is_file()):
        empty_dict: Dict[str, Any] = {}
        try:
            with open(path, "w") as wf:
                json.dump(empty_dict, wf)
        except FileNotFoundError as err:
            raise RuntimeError(
                "You have to create the folder for the json first."
            ) from err
    with open(f_n) as rf:
        return json.load(rf)


def write_json(f_n: Union[PathLike, str], obj_dict: Dict[str, Any]):
    """Write the obj_dict to the given file (json encoded).

    Can be used to store persistent variables between calls to the same script.

    Args:
        f_n (Union[PathLike, str]): Path to the file where the json encoded
        data is to be stored.
        obj_dict (Dict[str, Any]): Data that is to be stored.
    """
    with open(f_n, "w") as wf:
        json.dump(obj_dict, wf)



import logging

import numpy as np
import pytest
import torch as th
from gymnasium import spaces

from releso.feature_extractor import CombinedExtractor, FeatureExtractor


def observation_space():
    return spaces.Box(low=0, high=255, shape=(3, 224, 224), dtype=np.uint8)


@pytest.mark.torch_test
@pytest.mark.parametrize(
    [
        "observation_space",
        "feature_dim",
        "without_linear",
        "network_type",
        "logger",
        "error",
    ],
    [
        (
            observation_space(),
            None,
            None,
            None,
            logging.getLogger("test"),
            False,
        ),
        (observation_space(), None, None, "mobilenet_v2", None, False),
        (
            observation_space(),
            None,
            None,
            "inceptionv3",
            logging.getLogger("test"),
            "Given network type -inceptionv3- unknown.",
        ),
        (observation_space(), None, None, "resnet18", None, False),
        (observation_space(), None, True, "resnet18", None, False),
    ],
)
def test_feature_extractor(
    observation_space,
    feature_dim,
    without_linear,
    network_type,
    logger,
    error,
):
    calling_dict = {
        "observation_space": observation_space,
    }
    if feature_dim is not None:
        calling_dict["feature_dim"] = feature_dim
    else:
        feature_dim = 128
    if without_linear is not None:
        calling_dict["without_linear"] = without_linear
    if network_type is not None:
        calling_dict["network_type"] = network_type
    if logger is not None:
        calling_dict["logger"] = logger
    if error:
        with pytest.raises(RuntimeError) as error_info:
            model = FeatureExtractor(**calling_dict)
        assert error in str(error_info.value)
        return
    model = FeatureExtractor(**calling_dict)
    result = model.forward(
        th.as_tensor(observation_space.sample()[None]).float()
    )
    if without_linear:
        assert model.features_dim == result.shape[1]
    else:
        assert model.features_dim == feature_dim == result.shape[1]


@pytest.mark.torch_test
@pytest.mark.parametrize(
    [
        "observation_space",
        "feature_dim",
        "without_linear",
        "network_type",
        "logger",
        "error",
    ],
    [
        (
            observation_space(),
            None,
            None,
            None,
            logging.getLogger("test"),
            False,
        ),
    ],
)
def test_combined_feature_extractor(
    observation_space,
    feature_dim,
    without_linear,
    network_type,
    logger,
    error,
):
    observation_space = spaces.Dict(
        {
            "image": observation_space,
            "vector": spaces.Box(low=-1, high=1, shape=(10,), dtype=float),
        }
    )

    calling_dict = {
        "observation_space": observation_space,
    }
    if feature_dim is not None:
        calling_dict["cnn_output_dim"] = feature_dim
    else:
        feature_dim = 128
    if without_linear is not None:
        calling_dict["without_linear"] = without_linear
    if network_type is not None:
        calling_dict["network_type"] = network_type
    if logger is not None:
        calling_dict["logger"] = logger
    if error:
        with pytest.raises(RuntimeError) as error_info:
            model = FeatureExtractor(**calling_dict)
        assert error in str(error_info.value)
        return
    model = CombinedExtractor(**calling_dict)
    _ = model.forward(
        {
            key: th.as_tensor(value[None]).float()
            for key, value in observation_space.sample().items()
        }
    )



"""Util module.

The util module holds additional files which are used in this framework/
library but have no main functionality components. These include plotting
functions, logging functionality and other helper functions.
"""



"""Definition of the action space.

File holds all classes which define the spline and with that also the action
definition of the problem.
"""
import copy
from typing import Any, Dict, List, Optional, Union

import numpy as np
from pydantic.class_validators import root_validator, validator
from pydantic.types import conint

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.shape_parameterization import ShapeDefinition, VariableLocation
from releso.util.logger import get_parser_logger

try:
    from splinepy import NURBS, BSpline
except ImportError as err:  # pragma: no cover
    from releso.util.module_import_raiser import ModuleImportRaiser

    BSpline = ModuleImportRaiser("splinepy", err)
    NURBS = ModuleImportRaiser("splinepy", err)


class SplineSpaceDimension(BaseModel):
    """Defines a single spline space dimension of the current spline.

    The dimension is a dimension of the parametric spline dimension.
    """

    #: Name of the spline dimension used for easier identification
    name: str
    #: number of control points in this dimension
    number_of_points: conint(ge=1)
    #: degree of the spline in this dimension
    degree: conint(ge=1)
    #: knot vector describing the knot intervals of the spline in the current
    #: dimension
    knot_vector: Optional[List[float]]

    @validator("knot_vector", always=True)
    @classmethod
    def validate_knot_vector(
        cls, v: Optional[List[float]], values: Dict[str, Any]
    ) -> List[float]:
        """Validator for knot_vector.

        If knot vector not given tries to make a default open knot vector.

        Args:
            v ([type]): Value to validate.
            values ([type]): Already validated values.
            field ([type]): Name of the field that is currently validated.

        Raises:
            ParserException: Emitted parser error.

        Returns:
            float: value of the validated value.
        """
        if (
            "number_of_points" in values.keys()
            and "degree" in values.keys()
            and "name" in values.keys()
        ):
            n_knots = values["number_of_points"] + values["degree"] + 1
        else:
            raise ParserException(
                "SplineSpaceDimension",
                "knot_vector",
                "During validation the "
                "prerequisite variables number_of_points and degree were not "
                "present.",
            )
        if isinstance(v, list):
            get_parser_logger().debug(
                f"The knot_vector for dimension {values['name']} is given."
            )
            if len(v) == n_knots:
                return v
            else:
                raise ParserException(
                    "SplineSpaceDimension",
                    "knot_vector",
                    f"The knot vector does not contain the correct number of "
                    f"items for an open knot vector definition. (is: {len(v)},"
                    f" should_be: {n_knots}).",
                )
        elif v is None:
            get_parser_logger().debug(
                f"The knot_vector for dimension {values['name']} is not given,"
                f" trying to generate one in the open format."
            )
            starting_ending = values["degree"] + 1
            middle = n_knots - (2 * starting_ending)
            if middle >= 0:
                knot_vec = list(
                    np.append(
                        np.append(
                            np.zeros(starting_ending - 1),
                            np.linspace(0, 1, middle + 2),
                        ),
                        np.ones(starting_ending - 1),
                    )
                )
            else:
                get_parser_logger().warning(
                    f"The knot vector is shorter {n_knots} than the length "
                    f"given by the open format {starting_ending * 2}. Knot "
                    "vector is created by adding the starting and ending "
                    "parts. The knot vector might be to long."
                )
                knot_vec = list(
                    np.array(
                        [np.zeros(starting_ending), np.ones(starting_ending)]
                    ).flatten()
                )
            return knot_vec

    def get_knot_vector(self) -> List[float]:
        """Function returns the Knot vector given in the object.

        Returns:
            List[float]: Direct return of the inner variable.
        """
        return self.knot_vector


class SplineDefinition(ShapeDefinition):
    """Defines the spline.

    Base class for the NURBS and B-Spline implementations.
    """

    #: Definition of the space dimensions of the spline
    space_dimensions: List[SplineSpaceDimension]
    # #: Non parametric spline dimensions is currently not used.
    # spline_dimension: conint(ge=1)
    #: control points of the spline.
    control_points: Optional[List[List[VariableLocation]]]

    @root_validator(pre=True)
    @classmethod
    def make_default_control_point_grid(
        cls, values: Dict[str, Any]
    ) -> List[List[VariableLocation]]:
        """Validator for control_point_variables.

        If value is None a equidistant grid of control points will be given
        with variability of half the space between the each control point.

        Args:
            v ([type]): Value to validate.
            values ([type]): Already validated values.
            field ([type]): Name of the field that is currently validated.

        Raises:
            ParserException: Emitted parser error.

        Returns:
            List[List[VariableLocation]]: Definition of the control_points
        """
        if not values.get("control_points"):
            if "space_dimensions" not in values.keys():
                raise ParserException(
                    "SplineDefinition",
                    "control_point_variables",
                    "During validation the prerequisite variable "
                    f"space_dimensions was not present. {str(values)}",
                )
            spline_dimensions = values["space_dimensions"]
            n_points_in_dim = [
                dim["number_of_points"] for dim in spline_dimensions
            ]
            n_points_in_dim.reverse()
            # this can be done in a smaller and faster footprint but for
            # readability and understanding
            dimension_lists = []
            dim_spacings = []
            # create value range in each dimension separately
            for n_p in n_points_in_dim:
                dimension_lists.append(list(np.linspace(0.0, 1.0, n_p)))
                dim_spacings.append(1.0 / ((n_p - 1) * 2))
            # create each control point for by concatenating each value in each
            # list with each value of all other lists
            save_location = str(values["save_location"])
            v = None
            for inner_dim, dim_spacing in zip(dimension_lists, dim_spacings):
                # first iteration the v vector must be initialized with the
                # first dimension vector
                if v is None:
                    v = []
                    for element in inner_dim:
                        if element == 0.0:
                            v.append(
                                VariableLocation(
                                    current_position=element,
                                    min_value=element,
                                    max_value=element + dim_spacing,
                                    save_location=save_location,
                                )
                            )
                        elif element == 1.0:
                            v.append(
                                VariableLocation(
                                    current_position=element,
                                    min_value=element - dim_spacing,
                                    max_value=element,
                                    save_location=save_location,
                                )
                            )
                        else:
                            v.append(
                                VariableLocation(
                                    current_position=element,
                                    min_value=element - dim_spacing,
                                    max_value=element + dim_spacing,
                                    save_location=save_location,
                                )
                            )
                # for each successive dimension for each existing element in v
                # each value in the new dimension must be added
                else:
                    temp_v = []
                    for current_list in v:
                        for element in inner_dim:
                            elem = copy.deepcopy(
                                current_list
                                if isinstance(current_list, list)
                                else [current_list]
                            )
                            if element == 0.0:
                                temp_v.append(
                                    [
                                        VariableLocation(
                                            current_position=element,
                                            min_value=element,
                                            max_value=element + dim_spacing,
                                            save_location=save_location,
                                        )
                                    ]
                                    + elem
                                )
                            elif element == 1.0:
                                temp_v.append(
                                    [
                                        VariableLocation(
                                            current_position=element,
                                            min_value=element - dim_spacing,
                                            max_value=element,
                                            save_location=save_location,
                                        )
                                    ]
                                    + elem
                                )
                            else:
                                temp_v.append(
                                    [
                                        VariableLocation(
                                            current_position=element,
                                            min_value=element - dim_spacing,
                                            max_value=element + dim_spacing,
                                            save_location=save_location,
                                        )
                                    ]
                                    + elem
                                )
                    v = temp_v
            if isinstance(v[0], list):
                values["control_points"] = v
            else:
                values["control_points"] = [[v_i] for v_i in v]
        return values

    def get_number_of_points(self) -> int:
        """Returns the number of points in the Spline.

        Currently the number of points in the spline is calculated by
        multiplying the number of points in each spline dimension.

        Returns:
            int: number of points in the spline
        """
        return np.prod(
            [dimension.number_of_points for dimension in self.space_dimensions]
        )


class BSplineDefinition(SplineDefinition):
    """Definition of the BSpline implementation of the ReLeSO Toolbox."""

    def get_shape(self) -> BSpline:
        """Creates the BSpline from the definition given by the json file.

        Returns:
            BSpline: given by the #degrees and knot_vector in each
            space_dimension and the current control points.
        """
        self.get_logger().debug("Creating BSpline.")
        self.get_logger().debug(
            f"With control_points: {self.get_parameter_values()}"
        )

        return BSpline(
            [space_dim.degree for space_dim in self.space_dimensions],
            [
                space_dim.get_knot_vector()
                for space_dim in self.space_dimensions
            ],
            self.get_parameter_values(),
        )


class NURBSDefinition(SplineDefinition):
    """Definition of a NURBS spline.

    Definition of the NURBS implementation of the ReLeSO Toolbox, in
    comparison to the B-Spline implementation only an additional weights
    vector is added.
    """

    #: weights for the NURBS Spline definition. Other parameters are part of
    #: the spline definition class. Can be fixed or changeable
    weights: List[Union[float, VariableLocation]]

    @validator("weights")
    @classmethod
    def validate_weights(
        cls,
        v: Optional[List[Union[float, VariableLocation]]],
        values: Dict[str, Any],
    ) -> List[Union[float, VariableLocation]]:
        """Validator for variable weights.

        Validate if the correct number of weights are present in the weight
        vector. If weight vector is not given weight vector should be all ones.

        Args:
            v (List[Union[float, VariableLocation]]): Value to validate
            values (Dict[str, Any]): Previously validated variables

        Raises:
            ParserException: Parser Error

        Returns:
            List[Union[float, VariableLocation]]: Filled weight vector.
        """
        n_cp = np.prod(
            [
                space_dim.number_of_points
                for space_dim in values["space_dimensions"]
            ]
        )
        if isinstance(v, list):
            if len(v) == n_cp:
                get_parser_logger().debug(
                    "Found correct number of weights in SplineDefinition."
                )
            else:
                raise ParserException(
                    "SplineDefinition NURBS",
                    "weights",
                    f"The length of the "
                    f"weight vector {len(v)} is not the same as the number of "
                    f"control_points {n_cp}.",
                )

        # actually not functional since this would create a race condition
        # between BSpline and NURBS, since the only difference is the weights
        # variable.
        # elif v is None:
        #     v = list(np.ones(n_cp))

        return v

    @validator("weights", each_item=True)
    @classmethod
    def convert_weights_into_variable_location(
        cls, v: Union[float, VariableLocation], values: Dict[str, Any]
    ) -> VariableLocation:
        """Validator for variable weights.

        Convert all float values in the weight vector into VariableLocations.
        So that these can also be used as actions.

        Args:
            v (List[Union[float, VariableLocation]]): Value to validate
            values (Dict[str, Any]): Previously validated variables

        Returns:
            List[VariableLocation]: Filled weight vector.
        """
        if isinstance(v, float):
            save_location = str(values["save_location"])
            return VariableLocation(
                current_position=v, save_location=save_location
            )
        elif type(v) is VariableLocation:
            return v
        # validated by pydantic
        # else:
        #     raise ParserException(
        #         "SplineDefinition NURBS",
        #         "weights",
        #         "The weight vector contains a value which is not a float "
        #         f"or a VariableLocation. The type is {type(v)}.",
        #     )

    def get_weights(self) -> List[float]:
        """Returns the weights of the NURBS spline.

        Returns:
            List[float]: List of weights.
        """
        return [weight.current_position for weight in self.weights]

    def get_shape(self) -> NURBS:
        """Creates a NURBSSpline from the definition given by the json file.

        Returns:
            NURBS: NURBSSpline
        """
        self.get_logger().debug("Creating NURBS.")

        return NURBS(
            [space_dim.degree for space_dim in self.space_dimensions],
            [
                space_dim.get_knot_vector()
                for space_dim in self.space_dimensions
            ],
            self.get_parameter_values()[:-1],
            self.get_weights(),
        )

    def get_actions(self) -> List[VariableLocation]:
        """Extends the control point actions with the weight actions.

        Returns:
            List[VariableLocation]: List of possible actions for this NURBS
            spline.
        """
        actions = super().get_actions()
        actions.extend(
            [variable for variable in self.weights if variable.is_action()]
        )
        # raise RuntimeError(f"Actions: {actions}")
        return actions


    def get_parameter_values(self) -> List[List[float]]:
        """Returns the current positions of all control points with weights as
        well.

        Returns:
            List[List[float]]: Positions of all control points.
        """
        control_points = super().get_parameter_values()
        control_points.append(
            [weight.current_position for weight in self.weights]
        )
        return control_points

    def reset(self) -> None:
        """Resets the spline to the original shape."""
        super().reset()
        for weight in self.weights:
            weight.reset()



"""Loads XNS binary/ mixd format files.

This file contains functions to read the binary files
created by the xns solver.

This file is not tested due to it being very specific to the mixd/xns use case.
"""

import os
from struct import unpack
from typing import Tuple

import numpy as np


def read_mixd_double(
    filename: str, col: int, element_size: int = 8, element: str = ">d"
) -> np.ndarray:
    """Read a double array from a mixd file.

    Author: Daniel Wolff (wolff@avt.rwth-aachen.de)

    Args:
        filename (string): filename / location of the mixd file
        col (int): number of columns of the integer array
        element_size (int, optional): Size of the element in bytes.
            Defaults to 8.
        element (string, optional): Format of the element.
            Defaults to ">d".

    Returns:
        retValues (numpy.ndarray): Double array read from the provided file
    """
    with open(filename, "rb") as f_handle:
        size = os.stat(filename).st_size
        if size % (col * element_size) != 0:
            raise RuntimeError("Not enough columns in the requested file!")

        rows = size // (element_size * col)
        ret_values = np.zeros((rows, col))
        for i in range(rows):
            for j in range(col):
                data = f_handle.read(element_size)
                ret_values[i, j] = unpack(element, data)[0]
    return ret_values
    # _read_mixd_double()


def load_mixd(
    mesh_file: str, mesh_dim: int, solution_file: str, ndof: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Read in the mesh and the solution of an XNS simulation.

    After reading in the XNS solution it is stored in the correct attributes
    namely in the `coordinates` and `solution` attributes

    Author: Daniel Wolff (wolff@avt.rwth-aachen.de)

    Args:
        mesh_file (string): filename / location of the mesh file `mxyz`
        mesh_dim (int): Number of spatial dimensions of the mesh
        solution_file (string): filename / location of the file containing
        the nodal solutions
        ndof (int): Number of degrees of freedom per node

    Returns:
        coordinates (numpy.ndarray): Array of points from the mixd mesh
        values (numpy.ndarray): Nodal solution values from the mixd file
    """
    coordinates = read_mixd_double(mesh_file, mesh_dim)
    solution = read_mixd_double(solution_file, ndof)
    return coordinates, solution



"""Classes used to parse the environment definition.

Files defines the environment used for parsing and also most function which
hold the functionality for the Reinforcement Learning environment are
defined here.
"""
import multiprocessing
import pathlib
from timeit import default_timer as timer
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import gymnasium
import numpy as np
from gymnasium import spaces
from pydantic import UUID4, conint
from pydantic.class_validators import validator
from pydantic.fields import PrivateAttr
from stable_baselines3.common.monitor import Monitor

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.geometry import GeometryTypes
from releso.gym_environment import GymEnvironment
from releso.spor import SPORList
from releso.util.logger import VerbosityLevel, get_parser_logger, set_up_logger
from releso.util.types import ObservationType


class MultiProcessing(BaseModel):
    """Define multiprocessing capability of environment.

    Defines if the Problem should use Multiprocessing and with how many
    cores the solver can work. Does not force Multiprocessing for example
    if the solver does not support it.
    """

    #: Maximal number of cores which can be used by the current environment.
    #: Multi-Environments will use multiple of these.
    number_of_cores: conint(ge=1) = 1


class Environment(BaseModel):
    """Parser class of which the environment is based.

    Parser Environment object is created by pydantic during the parsing of the
    json object defining the RL based Shape optimization. Each object can
    create a gym environment that represents the given problem.
    """

    #: defines if multi-processing can be used.
    multi_processing: Optional[MultiProcessing]
    #: definition of the Geometry
    geometry: GeometryTypes
    #: definition of the spor objects
    spor: SPORList
    #: maximal number of timesteps to run each episode for
    max_timesteps_in_episode: Optional[conint(ge=1)] = None
    #: whether or not to reset the environment if the geometry has not change
    #: after a step
    end_episode_on_geometry_not_changed: bool = False
    #: reward if episode is ended due to reaching max step in episode
    reward_on_geometry_not_changed: Optional[float] = None
    #: reward if episode is ended due to geometry not changed
    reward_on_episode_exceeds_max_timesteps: Optional[float] = None
    # #: periodically save the end result of the optimization T-junction use
    # #: case TODO the next few cases I see personally more in a separate SPOR
    # #: Step but I am not sure how this can be worked since it needs data from
    # #: other sources
    # save_good_episode_results: bool = False
    # #: periodically save the end result of the optimization converging
    # #: channel use case
    # save_random_good_episode_results: bool = False
    # #: also saves the mesh when an episode is saved with
    # #: save_random_good_episode_results, works only if named option is True
    # save_random_good_episode_mesh: bool = False

    # object variables
    #: id if the environment, important for multi-environment learning
    _id: Optional[UUID4] = PrivateAttr(default=None)
    #: if validation environment validation ids are stored here
    _validation_ids: Optional[List[float]] = PrivateAttr(default=None)
    #: id of the current validation id
    _current_validation_idx: Optional[int] = PrivateAttr(default=None)
    #: number of timesteps currently spend in the episode
    _timesteps_in_episode: Optional[int] = PrivateAttr(default=0)
    #: StepReturn values from last step
    _last_step_results: Dict[str, Any] = PrivateAttr(default={})
    #: Toggle to whether or not it is possible to flatten the observation
    #: space. If the observation space is flattened the agents feature
    #: extractor is more compact
    _flatten_observations: bool = PrivateAttr(default=False)

    # The following are all variables of validation steps that are currently
    #  not implemented. These commented steps are still present due to the
    #  possibility of them being added again in the future. The optimal
    #  solution would be to add them as optional SPOR steps. This would be
    #  difficult due to limited information in each spor step.
    ##
    # #: path where the mesh should be saved to for validation
    # _validation_base_mesh_path: Optional[str] = PrivateAttr(default=None)
    # #: How many validations were already evaluated
    # _validation_iteration: Optional[int] = PrivateAttr(default=0)
    # #: Please check validation definition to see what this does
    # _save_image_in_validation: bool = PrivateAttr(default=False)
    # #: Please check validation definition to see what this does
    # _validation_timestep: int = PrivateAttr(default=0)
    # #: At these values the validation results are saved
    # _result_values_to_save: List[float] = PrivateAttr(
    #     default=list([0.2, 0.6, 0.9, 1.4, 1.8])
    # )
    # #: Approximate validation episode can be slightly wrong
    # _approximate_episode: int = PrivateAttr(default=-1)
    # #: Number of validation results which are already exported please check
    # #: function body to see/set the limit
    # _number_exported: int = PrivateAttr(default=0)

    @validator("reward_on_geometry_not_changed", always=True)
    @classmethod
    def check_if_reward_given_if_geometry_not_change_episode_killer_activated(
        cls, value, values
    ) -> float:
        """Validator reward_on_geometry_not_changed.

        Checks that 1) if a reward is set, also the boolean value for the
        end_episode_on_geometry_not_changed is True. 2) If
        end_episode_on_geometry_not_changed is True a reward value is set.

        Args:
            value (float): value to validate
            values (Dict[str, Any]): previously validated values
                    (here end_episode_on_geometry_not_changed is important)

        Raises:
            ParserException: Error is thrown if one of the conditions is
            not met.

        Returns:
            float: reward for the specified occurrence.
        """
        # due to pydantic this first statement should never be able to happen
        if (
            "end_episode_on_geometry_not_changed" not in values
        ):  # pragma: no cover
            raise ParserException(
                "Environment",
                "reward_on_geometry_not_changed",
                "Could not find definition of parameter "
                "end_episode_on_geometry_not_changed, please defines this "
                "variable since otherwise this variable would have no "
                "function.",
            )
        if value is not None and (
            values["end_episode_on_geometry_not_changed"] is None
            or not values["end_episode_on_geometry_not_changed"]
        ):
            raise ParserException(
                "Environment",
                "reward_on_geometry_not_changed",
                "Reward can only be set if end_episode_on_geometry_not_changed"
                " is true.",
            )
        if values["end_episode_on_geometry_not_changed"] and value is None:
            get_parser_logger().warning(
                "Please set a reward value for geometry not changed if episode"
                " should end on it. Will set 0 for you now, but this might "
                "not be you intention."
            )
            value = 0.0
        return value

    @validator("reward_on_episode_exceeds_max_timesteps", always=True)
    @classmethod
    def check_if_reward_given_if_max_steps_killer_activated(
        cls, value, values
    ) -> int:
        """Validator reward_on_episode_exceeds_max_timesteps.

        Checks that 1) if a reward is set, also the boolean value for the
        max_timesteps_in_episode is True. 2) If max_timesteps_in_episode
        is True a reward value is set.

        Args:
            value (float): value to validate
            values (Dict[str, Any]): previously validated values (here
            max_timesteps_in_episode is important)

        Raises:
            ParserException: Error is thrown if one of the conditions is
            not met.

        Returns:
            float: reward for the specified occurrence.
        """
        # due to pydantic this first statement should never be able to happen
        if "max_timesteps_in_episode" not in values:  # pragma: no cover
            raise ParserException(
                "Environment",
                "reward_on_episode_exceeds_max_timesteps",
                "Could not find definition of parameter "
                "max_timesteps_in_episode, please defines this variable since "
                "otherwise this variable would have no function.",
            )
        if value is not None and (
            values["max_timesteps_in_episode"] is None
            or not values["max_timesteps_in_episode"]
            or values["max_timesteps_in_episode"] <= 0
        ):
            raise ParserException(
                "Environment",
                "reward_on_episode_exceeds_max_timesteps",
                "Reward can only be set if max_timesteps_in_episode a positive"
                " integer.",
            )
        if values["max_timesteps_in_episode"] and value is None:
            get_parser_logger().warning(
                "Please set a reward value for max time steps exceeded, if "
                "episode should end on it. Will set 0 for you now, but this "
                "might not be your intention."
            )
            value = 0.0
        return value

    # object functions

    def __init__(self, **data: Any) -> None:
        """Construct the object."""
        super().__init__(**data)

    def _compress_observation_space_definition(
        self,
        observation_spaces: List[Tuple[str, ObservationType]],
        has_cnn_observations: bool,
    ) -> List[Tuple[str, ObservationType]]:
        """If possible compress observation space.

        If possible will compress the observation space into a single
        :py:`gymnasium.spaces.Box` observation space. This is not possible if
        cnn observations are being used and als not if the shape of the
        observations change between observations.

        Args:
            observation_spaces (List[Tuple[str,ObservationType]]):
                Uncompressed observations spaces.
            has_cnn_observations (bool): Observations include cnn observations

        Returns:
            List[Tuple[str,ObservationType]]: If possible compressed
                observation space if not unchanged observations space tuple
                list.
        """
        # check if there is potential for compressing the observation space
        if len(observation_spaces) > 1 and not has_cnn_observations:
            self.get_logger().info(
                "Found potential for compressed observation space."
            )
            new_space_min = []
            new_space_max = []
            new_space_shape = []
            for _, subspace in observation_spaces:
                new_space_min.extend(subspace.low)
                new_space_max.extend(subspace.high)
                new_space_shape.append(subspace.shape)
                if len(new_space_shape) > 1:
                    # if the shape do not match compression is not possible
                    if len(new_space_shape[-1]) != len(new_space_shape[-2]):
                        self.get_logger().info(
                            "Could not flatten observation space dict "
                            "definition into a single observation "
                            "space. Keeping dict definition."
                        )
                        break
                    else:
                        # adding up shapes
                        new_space_shape[0] = [
                            x + y
                            for x, y in zip(
                                new_space_shape[-1], new_space_shape[-2]
                            )
                        ]
                        new_space_shape.pop()
            else:
                # for loop completed so compression is possible
                self._flatten_observations = True
                observation_spaces = [
                    (
                        "flattened_observation_space",
                        spaces.Box(
                            low=np.array(new_space_min),
                            high=np.array(new_space_max),
                            shape=new_space_shape[0],
                        ),
                    )
                ]
        else:
            self.get_logger().info(
                "Did not find any potential for a "
                "compressed observation space."
            )

        return observation_spaces

    def _define_observation_space(self) -> gymnasium.Space:
        """Define the observation space of the environment.

        Creates the observation space the gym environment uses to define
        its observations space.

        Returns:
            # gymnasium.Space: Observation space of the current problem.
        """
        observation_spaces: List[Tuple[str, gymnasium.Space]] = []
        # define base observation
        if self.geometry.action_based_observation:
            observation_spaces.append(
                self.geometry.get_observation_definition()
            )
        # define spor observations
        has_cnn_observations = False
        spor_obs = self.spor.get_observations()
        if spor_obs is not None:
            for item in spor_obs:
                observation_spaces.append(item.get_observation_definition())
                if (
                    "value_type" in item.__dict__.keys()
                    and item.value_type == "CNN"
                ):
                    has_cnn_observations = True

        # check if dict is actually necessary
        observation_spaces = self._compress_observation_space_definition(
            observation_spaces, has_cnn_observations
        )

        if len(observation_spaces) > 1:
            # Observation space dict necessary
            observation_dict = dict(observation_spaces)
            self.get_logger().info(
                "Observation space is of type Dict and"
                " has the following description:"
            )
            for name, subspace in observation_spaces:
                self.get_logger().info(f"{name} has shape {subspace.shape}")
            return spaces.Dict(observation_dict)
        elif len(observation_spaces) == 1:
            # single observation space necessary
            self.get_logger().info(
                f"Observation space is NOT of type Dict and has the following"
                f" description: {observation_spaces[0][1].shape}"
            )
            # no dict space is needed so only the base observation space is
            # returned without a name
            return observation_spaces[0][1]
        else:
            raise RuntimeError(
                "The observation space is empty. Please define "
                "observations for RL to work."
            )

    def is_multiprocessing(self) -> int:
        """Check if environment uses multiprocessing.

        Function checks if the environment is setup to be used with
        multiprocessing Solver. Returns the number of cores the solver should
        use. If no multiprocessing 1 core is returned.

        Returns:
            int: Number of cores used in multiprocessing. 1 If no
                 multiprocessing. (Single thread still ok.)
        """
        return self.multi_processing.number_of_cores

    def get_validation_id(self) -> Optional[int]:
        """Return the validation id of the current run, if applicable.

        Checks if current environment has validation values if return the
        correct one otherwise return None.

        Returns:
            Optional[int]: Check text above.
        """
        if self._validation_ids is not None:
            return self._validation_ids[
                self._current_validation_idx % len(self._validation_ids)
            ]
        return None

    def step(self, action: Any) -> Tuple[Any, float, bool, Dict[str, Any]]:
        """Performs the step of the environment.

        Function that is called for each step. Contains all steps that are
        performed during each step inside the environment.

        There was a change of what the step function returns. It now returns
        the following values:

        1. observation
        2. reward
        3. terminated
        4. truncated
        5. info

        The change was that done was split into terminated and truncated.
        Terminated is now True if the episode is done. Truncated is True if
        the episode was ended due to the maximum number of steps, gone outside
        of physical bounds or action values.

        This change is really interesting for us as it means that we might be
        able to handle the termination of the episode other than goal_states
        better. Currently truncated is set to False to make it compatible.
        Needs work to be put in to allow this information to propagate
        correctly.

        Args:
            action (Any): Action value depends on if the ActionSpace is
            discrete (int - Signifier of the action) or
            Continuous (List[float] - Value for each continuous
            variable.)

        Returns:
            Tuple[Any, float, bool, bool, Dict[str, Any]]: [description]
        """
        start = timer()

        observations = {}
        done = False
        reward = 0.0
        info = {}

        # apply new action
        self.get_logger().info(f"Action {action}")
        info["geometry_information"] = self.geometry.apply_action(action)

        # run SPOR
        observations, reward, done, info = self.spor.run(
            step_information=(observations, done, reward, info),
            validation_id=self.get_validation_id(),
            core_count=self.is_multiprocessing(),
            reset=False,
            environment_id=self._id,
        )

        # check if geometry has not changed. But only in validation phase to
        #  exit episodes that are always repeating the same action without
        #  breaking.
        if not done:
            self._timesteps_in_episode += 1
            if (
                self.max_timesteps_in_episode
                and self.max_timesteps_in_episode > 0
                and self._timesteps_in_episode >= self.max_timesteps_in_episode
            ):
                done = True
                reward += self.reward_on_episode_exceeds_max_timesteps
                info["reset_reason"] = "max_timesteps_exceeded"
            if self.end_episode_on_geometry_not_changed:
                if not self.geometry.is_geometry_changed():
                    self.get_logger().info(
                        "The geometry observation have"
                        " not changed will exit episode."
                    )
                    reward += self.reward_on_geometry_not_changed
                    done = True
                    info["reset_reason"] = "geometry_not_changed"
        # else:
        #     pass  # TODO convert to separate SPOR Steps
        #     if reward >= 5.:
        #         if self.save_good_episode_results:
        #             for elem in self._result_values_to_save:
        #                 if np.isclose(
        #                     list(observations.values())[0][1],
        #                     elem, atol=0.05
        #                 ):
        #                     self.save_current_solution_as_png(
        #                         self.save_location/"episode_end_results"
        #                         ""/str(elem)/f""
        #                         f"{list(observations.values())[0][1]}"
        #                         f"_{self._approximate_episode}.png",
        #                         height=2, width=2, dpi=400)
        #                     break
        #         if self.save_random_good_episode_results:
        #             if self._number_exported < 200:
        #                 if np.random.default_rng().random() < 5:
        #                     self.save_current_solution_as_png(
        #                         self.save_location/"episode_end_results" /
        #                         f"{self._approximate_episode}.png",
        #                         height=2, width=2, dpi=400)
        #                     if self.save_random_good_episode_mesh:
        #                         self.export_mesh(
        #                             self.save_location/"episode_end_results"/
        #                             "mesh"/
        #                             f"{self._approximate_episode}.xns")
        #                     self._number_exported += 1

        self.get_logger().info(
            f"Current reward {reward} and episode is done: {done}."
        )

        self._last_step_results = {
            "observations": observations,
            "reward": reward,
            "done": done,
            "info": info,
        }
        self.get_logger().debug(self._last_step_results)
        # TODO convert to a SPOR Step
        # if self._validation_ids and self._save_image_in_validation:
        #     self._validation_timestep += 1
        #     self.save_current_solution_as_png(
        #         self.save_location/"validation"
        #         / str(self._validation_iteration)
        #         / str(self._current_validation_idx)
        #         / f"{self._validation_timestep}.png")

        observations["geometry_observation"] = self.geometry.get_observation()
        if observations["geometry_observation"] is None:
            del observations["geometry_observation"]
        end = timer()
        self.get_logger().debug(f"Step took {end - start} seconds.")

        return self.check_observations(observations), reward, done, False, info

    def check_observations(
        self, observations: ObservationType
    ) -> ObservationType:
        """Processes the observations.

        Transforms the observations from a dict observation to a normal
        observation if applicable.

        Args:
            observations (ObservationType): Observations to be processed.

        Returns:
            ObservationType: Processed Observations.
        """
        self.get_logger().debug(
            f"The observations are as follows: {observations}"
        )
        new_observation = []
        if len(observations.keys()) == 1:
            new_observation = observations[next(iter(observations.keys()))]
        else:
            if self._flatten_observations:
                for key in observations.keys():
                    new_observation.extend(observations[key])
            else:
                new_observation = observations
        return new_observation

    def reset(self, seed: Optional[int] = None) -> ObservationType:
        """Resets the environment.

        This can either be the case if the episode is done due to #time_steps
        or the environment emits the done signal.

        Args:
            seed (Optional[int], optional): Seed for the environment. Defaults
            to None.

        Returns:
            Tuple[Any]: Observation of the newly reset environment.
        """
        self.get_logger().info("Resetting the Environment.")

        observations = {}
        done = False
        reward = 0.0
        info = {}

        # reset geometry
        info["geometry_information"] = self.geometry.reset(
            self.get_validation_id()
        )

        self._timesteps_in_episode = 0
        # self._approximate_episode += 1
        # run solver and reset reward and get new solver observations
        observations, reward, done, info = self.spor.run(
            step_information=(observations, done, reward, info),
            validation_id=self.get_validation_id(),
            core_count=self.is_multiprocessing(),
            reset=True,
            environment_id=self._id,
        )

        if self._validation_ids:
            # TODO move to SPOR step
            # export mesh at end of validation
            # if self._current_validation_idx > 0 and \
            #         self._validation_base_mesh_path:
            #     # validation is performed in single environment
            #     # with no multi threading so this is not necessary.
            #     base_path = pathlib.Path(
            #         self._validation_base_mesh_path).parents[0]/str(
            #         self._validation_iteration)/str(
            #         self._current_validation_idx)
            #    file_name = pathlib.Path(self._validation_base_mesh_path).name
            #    if "_." in self._validation_base_mesh_path:
            #        validation_mesh_path = base_path / str(file_name).replace(
            #            "_.", f"{self.get_reset_reason_string()}.")
            #    else:
            #        validation_mesh_path = self._validation_base_mesh_path
            #    self.export_mesh(validation_mesh_path)
            #    self.export_spline(validation_mesh_path.with_suffix(".xml"))
            if self._current_validation_idx >= len(self._validation_ids):
                self.get_logger().info(
                    "The validation callback resets the environment one time "
                    "to often. Next goal state will again be the correct one."
                )
            self._current_validation_idx += 1
            if self._current_validation_idx > len(self._validation_ids):
                self._current_validation_idx = 0
                # self._validation_iteration += 1
        self.get_logger().info("Resetting the Environment DONE.")
        # TODO move to separate SPOR Step
        # if self._validation_ids and self._save_image_in_validation:
        #     self._validation_timestep = 0
        #     self.save_current_solution_as_png(
        #         self.save_location/"validation" /
        #         str(self._validation_iteration) /
        #         str(self._current_validation_idx) /
        #         f"{self._validation_timestep}.png")

        observations["geometry_observation"] = self.geometry.get_observation()
        if observations["geometry_observation"] is None:
            del observations["geometry_observation"]
        obs = self.check_observations(observations)
        return obs, info

    def set_validation(
        self,
        validation_values: List[float],
        end_episode_on_geometry_not_changed: bool = False,
        max_timesteps_in_episode: int = 0,
        reward_on_geometry_not_changed: Optional[float] = None,
        reward_on_episode_exceeds_max_timesteps: Optional[float] = None,
    ):
        """Converts the environment to a validation environment.

        This environment now only sets the goal states to the predefined
        values.

        Args:
            validation_values (List[float]): List of predefined goal states.
                base_mesh_path (Optional[str], optional): Path to the initial
                mesh. Defaults to None.
            end_episode_on_geometry_not_changed (bool, optional): Should the
                episode end if the geometry has no changes from one episode to
                the next. Defaults to False.
            max_timesteps_in_episode (int, optional): Maximal timesteps per
                episode, if 0 no limit. Defaults to 0.
            reward_on_geometry_not_changed (float, optional): Reward to give if
                episode is terminated due to unchanged geometry.
                Defaults to None.
            reward_on_episode_exceeds_max_timesteps(float, optional): Reward
                given of the episode is terminated due to exceeding the
                max_timesteps_in_episode. Defaults to None.
            save_image_in_validation (bool, optional): Should the validation
                save periodically the geometry/visualization. Is broken.
                Defaults to False.
        """
        self._validation_ids = validation_values
        self._current_validation_idx = 0
        # self._validation_base_mesh_path = base_mesh_path
        self.max_timesteps_in_episode = max_timesteps_in_episode
        self.end_episode_on_geometry_not_changed = (
            end_episode_on_geometry_not_changed
        )
        self.reward_on_geometry_not_changed = reward_on_geometry_not_changed
        self.reward_on_episode_exceeds_max_timesteps = (
            reward_on_episode_exceeds_max_timesteps
        )
        # self._save_image_in_validation = save_image_in_validation
        self.get_logger().info(
            f"Setting environment to validation. "
            f"max_timesteps {self.max_timesteps_in_episode}, "
            f"spine_not_changed {self.end_episode_on_geometry_not_changed}"
        )

    def get_gym_environment(
        self,
        logging_information: Optional[
            Dict[str, Union[str, pathlib.Path, VerbosityLevel]]
        ] = None,
    ) -> gymnasium.Env:
        """Creates the parametrized gymnasium environment.

        Creates and configures the gymnasium environment so it can be used
        for training.

        Returns:
            gymnasium.Env: OpenAI gymnasium environment that can be used to
                train with stable_baselines[3] agents.
        """
        if logging_information:
            self._set_up_logger(**logging_information)
        self.get_logger().info("Setting up Gym environment.")
        if self._id is None:
            self._id = uuid4()
        else:
            self.get_logger().warning(
                f"During setup of gym environment: The id of the environment "
                f"is already set to {self._id}. Please note that when using "
                "multi environment training this will lead to errors. Please "
                "use this function only after multiplying the environments "
                "and not before."
            )

        self.geometry.setup(self._id)
        env = GymEnvironment(
            self.geometry.get_action_definition(),
            self._define_observation_space(),
        )
        env.step = self.step
        env.reset = self.reset
        env.close = self.close
        return Monitor(env)

    def _set_up_logger(
        self,
        logger_name: str,
        log_file_location: pathlib.Path,
        logging_level: VerbosityLevel,
    ):
        """Setup the logger to be multiprocessing aware.

        Additional set up logger function for the purpose of multiprocessing
        capable logger initialization.

        Args:
            logger_name (str): name of the logger to set up
            log_file_location (pathlib.Path): location of the file where the
                logger should write to
            logging_level (VerbosityLevel): logger level to write up to
        """
        logger = multiprocessing.get_logger()
        resulting_logger = set_up_logger(
            logger_name, log_file_location, logging_level, logger=logger
        )
        self.set_logger_name_recursively(resulting_logger.name)

    def close(self):
        """Function is called when training is stopped."""



import copy
import pathlib
import shutil
import uuid
from collections.abc import Iterable
from typing import List

import numpy as np
import pytest
from conftest import dir_save_location_path
from pydantic import ValidationError

from releso.observation import (
    ObservationDefinition,
    ObservationDefinitionMulti,
)
from releso.spor import (  # SPORObjectPythonFunction,; SPORObjectTypes,
    MPIClusterMultiProcessor,
    MultiProcessor,
    SPORList,
    SPORObject,
    SPORObjectCommandLine,
    SPORObjectExecutor,
    SPORObjectExternalPythonFunction,
    SPORObjectInternalPythonFunction,
)
from releso.verbosity import VerbosityLevel


def get_obs_name(object_to_check) -> List[str]:
    if isinstance(
        object_to_check, (ObservationDefinition, ObservationDefinitionMulti)
    ):
        return [object_to_check.name]
    elif isinstance(object_to_check, Iterable):
        return [elem.name for elem in object_to_check]
    elif object_to_check is None:
        return []
    else:
        raise RuntimeError(f"{type(object_to_check)}, object_to_check")


@pytest.mark.parametrize(
    [
        "command",
        "max_core_count",
        "external_core_count",
        "always_use",
        "resulting_count",
    ],
    [
        (None, 1, 1, None, 1),
        (None, 10, 1, None, 1),
        (None, 10, 10, None, 10),
        (None, 1, 10, False, 1),
        (None, 10, -5, False, 1),
        (None, 1, 1, True, 1),
        (None, 10, 1, True, 1),
        (None, 10, 10, True, 10),
        (None, 1, 10, True, 1),
        (None, 10, -5, True, 1),
        ("this_is_a_command", 2, 2, True, 2),
    ],
)
def test_multi_processor(
    command,
    max_core_count,
    external_core_count,
    always_use,
    resulting_count,
    dir_save_location,
):
    calling_dict = {
        "save_location": dir_save_location,
        "max_core_count": max_core_count,
    }
    if command is not None:
        calling_dict["command"] = command
    if always_use is not None:
        calling_dict["always_use"] = always_use
    mp = MultiProcessor(**calling_dict)

    ret_command = mp.get_command(external_core_count)

    if resulting_count == 1 and not always_use:
        assert ret_command == ""
    else:
        if command is None:
            assert ret_command == f"mpiexec -n {resulting_count}"
        else:
            assert ret_command == f"{command} {resulting_count}"


@pytest.mark.parametrize(
    [
        "command",
        "command_variable",
        "mpi_flags_variable",
        "mpi_flags_variable_variable",
        "max_core_count",
        "external_core_count",
        "always_use",
        "resulting_command",
        "error",
    ],
    [
        (None, "mpiexec", None, "-np 2", 1, 1, True, "$MPIEXEC -np 1", False),
        (None, "mpiexec", None, "-np 2", 1, 1, False, "", False),
        (
            None,
            "total_annihilation",
            None,
            "",
            10,
            10,
            None,
            "$MPIEXEC -n 10",
            False,
        ),
        (
            "$ANOTHER_COMMAND",
            "simple_times",
            None,
            "los 3",
            5,
            -1,
            True,
            "$ANOTHER_COMMAND los 1",
            False,
        ),
        (
            "$ANOTHER_COMMAND",
            "simple_times",
            None,
            "los as 3",
            5,
            -1,
            True,
            "$ANOTHER_COMMAND los 1",
            "MPI cluster environment variable not as expected.",
        ),
        # (None, "mpiexec", 10, 10, None, 10),
        # (None, "mpiexec", 10, -5, False, 1),
        # (None, "mpiexec", 10, 1, True, 1),
        # (None, "mpiexec", 10, -5, True, 1),
        # ("$this_is_a_command", 2, 2, True, 2),
    ],
)
def test_mpi_cluster_multi_processor(
    command,
    command_variable,
    mpi_flags_variable,
    mpi_flags_variable_variable,
    max_core_count,
    external_core_count,
    always_use,
    resulting_command,
    error,
    dir_save_location,
    monkeypatch,
):
    calling_dict = {
        "save_location": dir_save_location,
        "location": "cluster",
        "max_core_count": max_core_count,
    }
    if always_use is not None:
        calling_dict["always_use"] = always_use
    if command is not None:
        calling_dict["command"] = command
    if command_variable is not None:
        monkeypatch.setenv(
            command[1:] if command else "MPIEXEC", command_variable
        )
    if mpi_flags_variable is not None:
        calling_dict["mpi_flags_variable"] = mpi_flags_variable
    if mpi_flags_variable_variable is not None:
        monkeypatch.setenv(
            mpi_flags_variable[1:]
            if mpi_flags_variable
            else "FLAGS_MPI_BATCH",
            mpi_flags_variable_variable,
        )
    mp = MPIClusterMultiProcessor(**calling_dict)
    mp.get_logger().handlers.clear()
    if error:
        with pytest.raises(RuntimeError):
            ret_command = mp.get_command(external_core_count)
        return
    ret_command = mp.get_command(external_core_count)
    assert ret_command == resulting_command


@pytest.mark.parametrize(
    "additional_observations, obs_type",
    [
        (12, ObservationDefinition),
        (str(0), None),
        (
            {"name": "test", "value_min": 0.0, "value_max": 1.0},
            ObservationDefinition,
        ),
        (
            [
                {"name": "test", "value_min": 0.0, "value_max": 1.0},
                ({"name": "test", "value_min": 0.0, "value_max": 1.0}),
            ],
            (list, ObservationDefinition),
        ),
    ],
)
def test_spor_object(
    additional_observations, obs_type, dir_save_location, caplog
):
    # get_parser_logger().handlers.clear()
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
    }
    if additional_observations:
        calling_dict["additional_observations"] = additional_observations
    with caplog.at_level(VerbosityLevel.WARNING):
        spor_object = SPORObject(**calling_dict)
    # We might want to deprecate the hidden option to instantiate additional
    # observations via an int. Building up a default observation definition.
    # if isinstance(additional_observations, int):
    #     assert (
    #         f"Please do not use this method to register additional observati"
    #         in caplog.text
    #     )
    obs = spor_object.get_observations()
    # check if observations are of correct type
    if isinstance(obs_type, tuple):
        # only for 2 element tuple like (list, ObservationDefinition)
        # only checks the first element of the obs if it is iterable
        element_to_check = obs
        for obs_type_item in obs_type:
            if obs_type_item is list:
                assert isinstance(element_to_check, obs_type_item)
                element_to_check = element_to_check[0]
            else:
                assert isinstance(element_to_check, obs_type_item)
    elif obs_type is None:
        assert obs is None
    else:
        assert issubclass(obs.__class__, obs_type)
    existing_obs = {}
    existing_obs_copy = copy.deepcopy(existing_obs)
    default_obs = spor_object.get_default_observation(existing_obs)
    if obs_type is None:
        assert default_obs == existing_obs
    else:
        assert default_obs != existing_obs_copy


@pytest.mark.parametrize(
    [
        "multi_processor",
        "wanted_mp",
        "use_communication_interface",
        "add_step_information",
        "working_dir",
        "error",
    ],
    [
        (
            {"max_core_count": 12},
            MultiProcessor,
            None,
            None,
            False,
            "does not exist or is not a valid directory",
        ),
        (
            {"max_core_count": 12},
            MultiProcessor,
            None,
            None,
            True,
            False,
        ),
        (
            {"max_core_count": 12},
            MultiProcessor,
            None,
            True,
            True,
            "Please only set the add_step_information variable to True",
        ),
        ({"max_core_count": 12}, MultiProcessor, True, True, True, False),
        (
            {"max_core_count": 12, "location": "cluster"},
            MPIClusterMultiProcessor,
            True,
            True,
            True,
            False,
        ),
    ],
)
def test_spor_object_executor(
    multi_processor,
    wanted_mp,
    use_communication_interface,
    add_step_information,
    working_dir,
    error,
    dir_save_location,
):
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
    }
    if working_dir:
        calling_dict["working_directory"] = str(dir_save_location_path(True))
    else:
        calling_dict["working_directory"] = str(dir_save_location_path())
    if multi_processor:
        multi_processor["save_location"] = dir_save_location
        calling_dict["multi_processor"] = multi_processor
    if use_communication_interface:
        calling_dict["use_communication_interface"] = (
            use_communication_interface
        )
    if add_step_information:
        calling_dict["add_step_information"] = add_step_information
    if error:
        with pytest.raises(ValidationError) as err:
            spor_object = SPORObjectExecutor(**calling_dict)
        assert error in str(err.value)
        return
    spor_object = SPORObjectExecutor(**calling_dict)
    assert type(spor_object.multi_processor) is wanted_mp


@pytest.mark.parametrize(
    "working_dir, wanted_result, relative, pre_create, error",
    [
        (".", ".", True, False, False),
        (
            "{}/test",
            "test_save_location_please_delete/test/",
            False,
            False,
            False,
        ),
        ("non_existing/folder", "as/", False, False, True),
        ("existing/folder/", "existing/folder/", True, True, False),
        (
            "{}/test/{}",
            "test_save_location_please_delete/test/idididid/",
            False,
            False,
            False,
        ),
    ],
)
def test_spor_object_executor_working_dir(
    working_dir, wanted_result, relative, pre_create, error, dir_save_location
):
    if pre_create:
        pathlib.Path(working_dir).mkdir(parents=True, exist_ok=True)

    if not relative:
        wanted_result = str(
            (
                pathlib.Path("/".join(str(dir_save_location).split("/")[:-1]))
                / wanted_result
            )
            .resolve()
            .absolute()
        )

    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
        "multi_processor": {
            "max_core_count": 1,
            "save_location": dir_save_location,
        },
    }
    calling_dict["working_directory"] = str(working_dir)
    if error:
        with pytest.raises(ValidationError):
            spor_object = SPORObjectExecutor(**calling_dict)
            spor_object.setup_working_directory("idididid")
        return
    spor_object = SPORObjectExecutor(**calling_dict)
    spor_object.setup_working_directory("idididid")
    print(spor_object.working_directory)
    # if relative:
    #     assert spor_object.working_directory == wanted_result
    # else:
    assert spor_object.working_directory == wanted_result
    if pre_create:
        shutil.rmtree(wanted_result)


@pytest.mark.parametrize("with_multi_processing", [(True), (False)])
def test_spor_object_executor_setup_working_dir(
    with_multi_processing, dir_save_location, caplog, clean_up_provider
):
    dir_save_location_path(True)
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
        "multi_processor": {
            "max_core_count": 12,
            "save_location": dir_save_location,
        },
    }
    if with_multi_processing:
        calling_dict["working_directory"] = str(dir_save_location) + "/{}"
    else:
        calling_dict["working_directory"] = str(dir_save_location)

    spor_object = SPORObjectExecutor(**calling_dict)
    env_id = uuid.uuid4()
    with caplog.at_level(VerbosityLevel.INFO):
        spor_object.setup_working_directory(env_id)
    if with_multi_processing:
        assert "Found placeholder for step with name" in caplog.text
        assert (dir_save_location / str(env_id)).is_dir()
    else:
        assert "Found placeholder for step with name" not in caplog.text
        assert not (dir_save_location / str(env_id)).is_dir()
    clean_up_provider(dir_save_location)

    assert "" == spor_object.get_multiprocessing_prefix(1)
    assert "12" in spor_object.get_multiprocessing_prefix(12)


@pytest.mark.parametrize(
    "use_comm_interface, reset, validation_id, step_info, add_step_info",
    [
        (False, False, None, None, False),
        (True, False, None, None, False),
        (True, True, None, None, False),
        (True, False, 1234, None, False),
        (
            True,
            False,
            None,
            ["something", "something else", 1234, 0.1234],
            True,
        ),
        (
            True,
            True,
            1234,
            ["something", "something else", 1234, 0.1234],
            False,
        ),
    ],
)
def test_spor_object_executor_spor_com_interface(
    use_comm_interface,
    reset,
    validation_id,
    step_info,
    add_step_info,
    dir_save_location,
    caplog,
    clean_up_provider,
):
    dir_save_location_path(True)
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
        "multi_processor": {
            "max_core_count": 12,
            "save_location": dir_save_location,
        },
        "working_directory": str(dir_save_location),
        "use_communication_interface": use_comm_interface,
        "add_step_information": add_step_info,
    }
    spor_object = SPORObjectExecutor(**calling_dict)
    env_id = uuid.uuid4()
    func_calling_dict = {
        "environment_id": env_id,
        "step_information": step_info,
        "reset": reset,
        "validation_id": validation_id,
    }
    # raise RuntimeError(f"{step_info}")
    interface_string = spor_object.spor_com_interface(**func_calling_dict)
    if not use_comm_interface:
        assert len(interface_string) == 0
        return
    interface_string = " ".join(interface_string)
    if reset:
        assert "--reset" in interface_string
    else:
        assert "--reset" not in interface_string
    if validation_id is not None:
        assert f"--validation_value {str(validation_id)}" in interface_string
    else:
        assert "--validation_value" not in interface_string
    if add_step_info:
        assert "--json_object" in interface_string
    else:
        assert "--json_object" not in interface_string


@pytest.mark.parametrize(
    "observation, observation_addition, old_done, new_done, reset_reason",
    [
        (3, [1, 2, 3], False, False, None),
        (3, [1, 2, 3], False, False, "reset_reason"),
        (
            [
                {"name": "test1", "value_min": 0.0, "value_max": 1.0},
                {"name": "test2", "value_min": 0.0, "value_max": 1.0},
            ],
            {"test1": 0.5, "test2": 0.5},
            False,
            False,
            None,
        ),
        (None, None, False, False, None),
    ],
)
def test_spor_object_executor_interface_add(
    observation,
    observation_addition,
    old_done,
    new_done,
    reset_reason,
    dir_save_location,
):
    dir_save_location_path(True)
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
        "multi_processor": {
            "max_core_count": 12,
            "save_location": dir_save_location,
        },
        "working_directory": str(dir_save_location),
    }
    if observation is not None:
        calling_dict["additional_observations"] = observation
    spor_object = SPORObjectExecutor(**calling_dict)
    new_step_info = {
        "observations": observation_addition,
        "info": {spor_object.name: {"ret_info": 42}},
        "done": new_done,
        "reward": 0.0,
    }
    if reset_reason is not None:
        new_step_info["info"]["reset_reason"] = reset_reason
    # raise RuntimeError(f"{new_step_info}")
    step_info = {
        "observation": {},
        "info": {spor_object.name: {}},
        "done": new_done,
        "reward": -10.0,
    }

    spor_object.spor_com_interface_add(
        returned_step_dict=new_step_info, step_dict=step_info
    )
    if reset_reason is not None:
        assert reset_reason == step_info["info"]["reset_reason"]
    assert (old_done or new_done) == step_info["done"]
    assert step_info["reward"] == 0.0


@pytest.mark.parametrize(
    [
        "observation",
        "observation_addition",
        "old_done",
        "new_done",
        "reset_reason",
        "error",
    ],
    [
        (3, [1, 2, 3], False, False, None, False),
        (3, [1, 2, 3], False, False, "reset_reason", False),
        (
            [
                {"name": "test1", "value_min": 0.0, "value_max": 1.0},
                {"name": "test2", "value_min": 0.0, "value_max": 1.0},
            ],
            {"test1": 0.5, "test2": 0.5},
            False,
            False,
            None,
            False,
        ),
        (None, None, False, False, None, False),
        (None, None, False, False, None, True),
    ],
)
def test_spor_object_executor_interface_read(
    observation,
    observation_addition,
    old_done,
    new_done,
    reset_reason,
    error,
    dir_save_location,
):
    dir_save_location_path(True)
    calling_dict = {
        "name": "test",
        "save_location": dir_save_location,
        "reward_on_error": 0.0,
        "multi_processor": {
            "max_core_count": 12,
            "save_location": dir_save_location,
        },
        "working_directory": str(dir_save_location),
    }
    if observation is not None:
        calling_dict["additional_observations"] = observation
    spor_object = SPORObjectExecutor(**calling_dict)
    new_step_info = {
        "observations": observation_addition,
        "info": {spor_object.name: {"ret_info": 42}},
        "done": new_done,
        "reward": 0.0,
    }
    if reset_reason is not None:
        new_step_info["info"]["reset_reason"] = reset_reason
    # raise RuntimeError(f"{new_step_info}")
    step_info = {
        "observation": {},
        "info": {spor_object.name: {}},
        "done": new_done,
        "reward": -10.0,
    }
    if error:
        with pytest.raises(SyntaxError):
            spor_object.spor_com_interface_read(
                output=bytes(str(new_step_info)[4:], "utf-8"),
                step_dict=step_info,
            )
        return
    b_ret_dict = bytes(str(new_step_info), "utf-8")
    spor_object.spor_com_interface_read(output=b_ret_dict, step_dict=step_info)
    if reset_reason is not None:
        assert reset_reason == step_info["info"]["reset_reason"]
    assert (old_done or new_done) == step_info["done"]
    assert step_info["reward"] == 0.0
    obs_names = get_obs_name(spor_object.additional_observations)
    if len(obs_names) == 1:
        obs_name = obs_names[0]
        assert np.allclose(
            step_info["observation"][obs_name],
            new_step_info["observations"],
        )
    else:
        for obs_name in obs_names:
            assert np.allclose(
                step_info["observation"][obs_name],
                new_step_info["observations"][obs_name],
            )


@pytest.mark.parametrize(
    [
        "reset",
        "run_on_reset",
        "additional_observations",
        "patch_function",
        "reward_on_completion",
    ],
    [
        (False, False, 3, False, None),
        (True, False, 3, False, None),
        (False, False, None, False, None),
        (True, True, None, False, None),
        (True, True, None, True, None),
        (True, True, None, True, 12),
    ],
)
def test_spor_internal_python_function(
    reset,
    run_on_reset,
    additional_observations,
    patch_function,
    reward_on_completion,
    dir_save_location,
    monkeypatch,
    caplog,
):
    dir_save_location_path(True)
    if patch_function:
        from releso.util import cnn_xns_observations

        def patch_function(*args, **kwargs):
            return {
                "reward": 2,
                "info": {},
                "done": False,
                "observations": np.zeros((3, 200, 200)),
            }, args[2]

        monkeypatch.setattr(cnn_xns_observations, "main", patch_function)

    calling_dict = {
        "name": "xns_cnn",
        "function_name": "xns_cnn",
        "save_location": dir_save_location,
        "working_directory": str(dir_save_location),
        "reward_on_error": 0.0,
        "use_communication_interface": True,
        "run_on_reset": run_on_reset,
    }
    if additional_observations is not None:
        calling_dict["additional_observations"] = additional_observations
    if reward_on_completion is not None:
        calling_dict["reward_on_completion"] = reward_on_completion
    spor_object = SPORObjectInternalPythonFunction(**calling_dict)

    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    with caplog.at_level(VerbosityLevel.INFO):
        observations, reward, done, info = spor_object.run(
            (observations, reward, done, info),
            environment_id=env_id,
            reset=reset,
        )
    if reward_on_completion is not None:
        assert reward == reward_on_completion
        assert (
            "The default reward will overwrite the already set reward"
            in caplog.text
        )


@pytest.mark.parametrize(
    "file_path, error_init, error_run",
    [
        (
            "/samples/spor_python_scripts_tests/file_exists_has_main.py",
            False,
            False,
        ),
        (
            "/samples/spor_python_scripts_tests/file_exists_no_main.py",
            False,
            "Could not get main function from python file",
        ),
        (
            "/samples/spor_python_scripts_tests/file_exists_unknown_import.py",
            False,
            "Could not load the python file",
        ),
        (
            "/samples/spor_python_scripts_tests/file_exists.py",
            "does not exist.",
            False,
        ),
    ],
)
def test_spor_external_python_function(
    file_path, error_init, error_run, dir_save_location, clean_up_provider
):
    dir_save_location_path(True)
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    calling_dict = {
        "name": "dummy_spor",
        "python_file_path": file_prefix + file_path,
        "save_location": dir_save_location,
        "working_directory": str(dir_save_location),
        "reward_on_error": 0.0,
        "use_communication_interface": True,
        "additional_observations": 3,
    }
    if error_init:
        with pytest.raises(ValidationError) as err:
            spor_obj = SPORObjectExternalPythonFunction(**calling_dict)
        assert error_init in str(err.value)

        clean_up_provider(dir_save_location)
        return
    spor_obj = SPORObjectExternalPythonFunction(**calling_dict)

    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    if error_run:
        with pytest.raises(RuntimeError) as err:
            spor_obj.run(
                (observations, reward, done, info), environment_id=env_id
            )
        assert error_run in str(err.value)
        clean_up_provider(dir_save_location)
        return
    observations, reward, done, info = spor_obj.run(
        (observations, reward, done, info), environment_id=env_id
    )
    assert reward == 2
    assert np.allclose(
        observations[spor_obj.additional_observations.name], [1, 2, 3]
    )
    clean_up_provider(dir_save_location)


def test_spor_external_python_function_reset_reason(dir_save_location):
    dir_save_location_path(True)
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    calling_dict = {
        "name": "dummy_spor",
        "python_file_path": (
            file_prefix
            + "/samples/spor_python_scripts_tests/file_exists_has_main.py"
        ),
        "save_location": dir_save_location,
        "working_directory": str(dir_save_location),
        "reward_on_error": 0.0,
        "use_communication_interface": True,
        "additional_observations": 3,
    }
    spor_obj = SPORObjectExternalPythonFunction(**calling_dict)

    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    while not done:
        observations, reward, done, info = spor_obj.run(
            (observations, reward, done, info), environment_id=env_id
        )
    assert done
    assert info[spor_obj.name]["reset_reason"] == "func_data > 10"
    assert info["reset_reason"] == "func_data > 10"


@pytest.mark.parametrize(
    [
        "execution_command",
        "error_init",
        "command_options",
        "error_run",
        "reward_on_completion",
        "reset",
    ],
    [
        (
            "python",
            None,
            "samples/spor_python_scripts_tests/file_exists_command_line.py",
            False,
            False,
            False,
        ),
        (
            "python",
            None,
            "samples/spor_python_scripts_tests/file_exists_command_line.py",
            False,
            True,
            False,
        ),
        (
            "python",
            None,
            "samples/spor_python_scripts_tests/file_exists_command_line.py",
            False,
            False,
            True,
        ),
        (
            "python",
            None,
            "samples/spor_python_scripts_tests/file_exists_command_l.py",
            "Some error",
            False,
            False,
        ),
        (
            str(dir_save_location_path(True)),
            "The execution_command path",
            False,
            False,
            False,
            False,
        ),
    ],
)
def test_spor_command_line(
    execution_command,
    error_init,
    command_options,
    error_run,
    reward_on_completion,
    reset,
    clean_up_provider,
    dir_save_location,
    caplog,
):
    dir_save_location_path(True)
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    calling_dict = {
        "name": "dummy_spor",
        "execution_command": execution_command,
        "save_location": dir_save_location,
        "working_directory": str(dir_save_location),
        "reward_on_error": 0.0,
        "use_communication_interface": True,
        "additional_observations": 3,
    }
    if command_options:
        calling_dict["command_options"] = [f"{file_prefix}/{command_options}"]
    if reward_on_completion:
        calling_dict["reward_on_completion"] = 12
    if reset:
        calling_dict["run_on_reset"] = False
    if error_init:
        with pytest.raises(ValidationError) as err:
            spor_obj = SPORObjectCommandLine(**calling_dict)
        assert error_init in str(err.value)

        clean_up_provider(dir_save_location)
        return
    spor_obj = SPORObjectCommandLine(**calling_dict)

    # error_run = False
    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    # create folder for step
    file_path = dir_save_location / f"{env_id}/{spor_obj._run_id}.json"
    file_path.parent.mkdir(parents=True, exist_ok=True)
    observations, reward, done, info = spor_obj.run(
        (observations, reward, done, info),
        environment_id=env_id,
        reset=reset,
    )
    if error_run and error_run is not None:
        # default observation
        assert np.allclose(
            observations[spor_obj.additional_observations.name], [-1, -1, -1]
        )
    elif reward_on_completion:
        assert reward == 12
        assert np.allclose(
            observations[spor_obj.additional_observations.name], [1, 2, 3]
        )
    elif reset:
        # default observation
        assert np.allclose(
            observations[spor_obj.additional_observations.name], [-1, -1, -1]
        )
        assert pytest.approx(0) == reward
    else:
        assert reward == 2
        assert np.allclose(
            observations[spor_obj.additional_observations.name], [1, 2, 3]
        )
    clean_up_provider(dir_save_location)


@pytest.mark.parametrize(
    [
        "rewards",
        "reward_aggregation",
        "change_aggregator",
        "w_sum",
        "error",
    ],
    [
        ([1, 2, 3, 4.0], "mean", None, 2.5, False),
        ([1, 2, 3, 4.0], "sum", None, 10, False),
        ([1, 2, 3, 4.0], "mean", "sum", 10, False),
        ([1, 2, 3, 4.0], "min", None, 1, False),
        ([-25, 1, 2, 3, 4.0], "max", None, 4, False),
        ([1, 2, 3, 4.0], "absmax", None, 4, False),
        ([-25, 1, 2, 3, 4.0], "absmax", None, -25, False),
        (
            [1, 2, 3, 4.0],
            "mean",
            "test",
            4,
            "The given reward aggregation method",
        ),
        ([], "absmax", None, 0, False),
    ],
)
def test_spor_list_compute_reward(
    dir_save_location,
    rewards,
    reward_aggregation,
    change_aggregator,
    w_sum,
    error,
):
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [],
        "reward_aggregation": reward_aggregation,
    }
    spor_list = SPORList(**calling_dict)
    spor_list._rewards = rewards
    if change_aggregator:
        spor_list.reward_aggregation = change_aggregator
    if error:
        with pytest.raises(RuntimeError) as err:
            spor_list._compute_reward()
        assert error in str(err.value)
        return
    reward = spor_list._compute_reward()
    assert pytest.approx(w_sum) == reward


def test_spor_list_get_observations_correct_spor_type(dir_save_location):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "additional_observations": [
                    {"name": "test1", "value_min": 0.0, "value_max": 1.0},
                    {"name": "test2", "value_min": 0.0, "value_max": 1.0},
                ],
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
            {
                "name": "internal",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "function_name": "xns_cnn",
            },
            {
                "name": "command_line",
                "save_location": dir_save_location,
                "execution_command": "python",
                "command_options": [
                    (
                        f"{file_prefix}/samples/spor_python_scripts_tests/"
                        "file_exists_command_line.py"
                    )
                ],
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
            },
        ],
        "reward_aggregation": "mean",
    }
    dir_save_location_path(True)
    spor_list = SPORList(**calling_dict)
    for idx, spor_obj in enumerate(spor_list.steps):
        if idx == 0:
            assert type(spor_obj) is SPORObjectExternalPythonFunction
            assert type(spor_obj) not in [
                SPORObjectInternalPythonFunction,
                SPORObjectCommandLine,
            ]
        elif idx == 1:
            assert type(spor_obj) is SPORObjectInternalPythonFunction
            assert type(spor_obj) not in [
                SPORObjectExternalPythonFunction,
                SPORObjectCommandLine,
            ]
        elif idx == 2:
            assert type(spor_obj) is SPORObjectCommandLine
            assert type(spor_obj) not in [
                SPORObjectExternalPythonFunction,
                SPORObjectInternalPythonFunction,
            ]
    obs = spor_list.get_observations()
    assert len(obs) == 3


def test_spor_list_observations_none(dir_save_location):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    dir_save_location_path(True)
    # check for None obs
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
        ],
        "reward_aggregation": "mean",
    }
    spor_list = SPORList(**calling_dict)
    assert spor_list.get_observations() is None


def test_spor_list_run_simple(dir_save_location, clean_up_provider):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
        ],
        "reward_aggregation": "mean",
    }
    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    spor_list = SPORList(**calling_dict)
    observations, reward, done, info = spor_list.run(
        (observations, reward, done, info),
        environment_id=env_id,
    )
    clean_up_provider(dir_save_location)


def test_spor_list_run_observations(dir_save_location, clean_up_provider):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "additional_observations": 3,
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
            {
                "name": "external1",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "additional_observations": 3,
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
            {
                "name": "external2",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "additional_observations": 3,
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
            {
                "name": "external3",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "additional_observations": 3,
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
        ],
        "reward_aggregation": "mean",
    }
    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    spor_list = SPORList(**calling_dict)
    observations, reward, done, info = spor_list.run(
        (observations, reward, done, info),
        environment_id=env_id,
    )
    obs = spor_list.get_observations()
    assert len(observations.keys()) == len(obs) == 4
    for ob in obs:
        obs_name = ob.name
        assert np.allclose(observations[obs_name], [1, 2, 3])

    clean_up_provider(dir_save_location)


def test_spor_list_run_done(dir_save_location, clean_up_provider):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
        ],
        "reward_aggregation": "mean",
    }
    env_id = uuid.uuid4()
    observations = {}
    done = True
    reward = 0.0
    info = {}
    spor_list = SPORList(**calling_dict)
    observations, reward, done, info = spor_list.run(
        (observations, reward, done, info),
        environment_id=env_id,
    )
    assert info.get("reset_reason") == "unspecified"
    clean_up_provider(dir_save_location)


def test_spor_list_run_reset(dir_save_location, clean_up_provider, caplog):
    file_prefix = str(pathlib.Path(__file__).parent.resolve())
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "steps": [
            {
                "name": "produces_error",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "working_directory": str(dir_save_location),
                "execution_command": "python",
                "command_options": [
                    str(dir_save_location / "does_not_exist.txt")
                ],
                "stop_after_error": True,
            },
            {
                "name": "external",
                "save_location": dir_save_location,
                "reward_on_error": 0.0,
                "use_communication_interface": True,
                "working_directory": str(dir_save_location),
                "run_on_reset": False,
                "additional_observations": 3,
                "python_file_path": (
                    f"{file_prefix}/samples/spor_python_scripts_tests/"
                    "file_exists_has_main.py"
                ),
            },
        ],
        "reward_aggregation": "mean",
    }
    env_id = uuid.uuid4()
    observations = {}
    done = False
    reward = 0.0
    info = {}
    spor_list = SPORList(**calling_dict)
    with caplog.at_level(VerbosityLevel.WARNING):
        observations, reward, done, info = spor_list.run(
            (observations, reward, done, info),
            environment_id=env_id,
            reset=True,
        )
    assert done

    clean_up_provider(dir_save_location)
    clean_up_provider(dir_save_location)



"""Framework experiment definition and experiment conduction.

File defines the base json object which is needed to define the problem setting
for the command line based usage of the ReLeSO toolbox/framework.
"""

import pathlib
from copy import deepcopy
from typing import Any, Optional, Union

import numpy as np
from pydantic.fields import Field, PrivateAttr
from pydantic.types import conint
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.callbacks import StopTrainingOnMaxEpisodes
from stable_baselines3.common.vec_env import DummyVecEnv, SubprocVecEnv, VecEnv

from releso.agent import AgentTypeDefinition
from releso.base_model import BaseModel
from releso.callback import EpisodeLogCallback, StepLogCallback
from releso.exceptions import ValidationNotSet
from releso.parser_environment import Environment
from releso.validation import Validation
from releso.verbosity import Verbosity


class BaseParser(BaseModel):
    """Class parses the experiment definition and conducts the training.

    This class can be used to initialize the ReLeSO Framework from the command
    line by reading in a json representation of the RL based Shape
    Optimization problem which is to be solved via Reinforcement Learning.
    """

    #: Defining the verbosity of the training process and environment loading
    verbosity: Verbosity = Field(default_factory=Verbosity)
    #: Definition of the agent to be used during training and/or validation of
    #: the RL use case
    agent: AgentTypeDefinition
    #: Definition of the environment which encodes the parameters of the RL use
    #: case
    environment: Environment
    #: Number of timesteps the training process should run for
    number_of_timesteps: conint(ge=1)
    #: Number of episodes the training process should run for. If given both
    #: timesteps and max episodes can stop the trainings progress.
    #: Default: None
    number_of_episodes: Optional[conint(ge=1)] = None
    #: Definition of the validation . Defaults to None.
    validation: Optional[Validation] = None
    #: Number of environments to train in parallel. Defaults to None.
    n_environments: Optional[conint(ge=1)] = 1
    #: Should training parameters be normalized to the number of environments?
    #: If True the number of steps between learnings are divided by the number
    #: of environments. This increases the training speed for PPO and A2C

    #: but the training might be a little bit more unstable.
    #: Defaults to False.
    normalize_training_values: bool = False
    #: Should the multi environment be run sequentially (True) or with multi
    #: processing (False). Defaults to False.
    multi_env_sequential: bool = False
    #: Number of episodes after which the episode_log is updated. It will be
    #: updated at the end of the training in any case. But making this number
    #: higher will lower the computational overhead. Defaults to 100.
    episode_log_update: conint(ge=1) = 100
    #: Flag indicating whether the step information (like actions,
    #: observations, ...) should be logged to file. Defaults to False.
    export_step_log: bool = False
    #: Number of steps after which the step_log is updated. It will be
    #: updated at the end of the training in any case. But making this number
    #: higher will lower the computational overhead. Defaults to 0 which
    #: triggers the output after every episode.
    step_log_update: conint(ge=0) = 0
    #: Flag indicating whether the step_log should also contain the
    #: information of the environment step. Defaults to False.
    step_log_info: bool = False

    # internal objects
    #: Holds the trainable agent for the RL use case. The
    #: ReLeSO.base_parser.BaseParser.agent defines the type and parameters of
    #: the agent this is the actual trainable agent.
    _agent: Optional[BaseAlgorithm] = PrivateAttr(default=None)

    def __init__(self, **data: Any) -> None:
        """Constructor of the base parser.

        Initializes the class correctly and also adds the correct logger to all
        subclasses.
        """
        super().__init__(**data)
        self.set_logger_name_recursively(self.verbosity._environment_logger)

    def learn(self) -> None:
        """Starts the training that is specified in the loaded json file.

        Successive calls to this function does not train the agent further

        but reinitialize the agent.
        """
        train_env: Optional[VecEnv] = None
        validation_environment = self._create_validation_environment()
        normalizer_divisor = (
            1
            if self.n_environments is None
            or not self.normalize_training_values
            else self.n_environments
        )
        if self.n_environments and self.n_environments > 1:
            env_create_list = [
                self._create_new_environment(
                    self.get_logger().name + f"_{idx}"
                )
                for idx in range(self.n_environments)
            ]
            if self.multi_env_sequential:
                train_env = DummyVecEnv(env_create_list)
            else:
                train_env = SubprocVecEnv(env_create_list)
        else:
            train_env = DummyVecEnv([
                lambda: self.environment.get_gym_environment()
            ])

        self._agent = self.agent.get_agent(train_env, normalizer_divisor)
        self.get_logger().info(f"Agent is of type {type(self._agent)}")
        callbacks = [
            EpisodeLogCallback(
                episode_log_location=self.save_location / "episode_log.csv",
                verbose=1,
                update_n_episodes=self.episode_log_update,
                logger=self.get_logger(),
            ),
        ]

        if self.export_step_log:
            callbacks.append(
                StepLogCallback(
                    step_log_location=self.save_location / "step_log.jsonl",
                    verbose=1,
                    update_n_steps=self.step_log_update,
                    log_infos=self.step_log_info,
                    logger=self.get_logger(),
                ),
            )

        if self.number_of_episodes is not None:
            num = self.number_of_episodes
            if self.normalize_training_values:
                num = int(num / normalizer_divisor)
            callbacks.append(StopTrainingOnMaxEpisodes(max_episodes=num))

        if self.validation:
            if self.validation.should_add_callback():
                callbacks.append(
                    self.validation.get_callback(
                        validation_environment.get_gym_environment(),
                        save_location=self.save_location,
                        normalizer_divisor=normalizer_divisor,
                    )
                )

        self.get_logger().info(
            f"The environment is now trained for {self.number_of_episodes} "
            f"episodes or {self.number_of_timesteps} timesteps."
        )
        self._agent.env.reset()
        self._agent.learn(
            self.number_of_timesteps,
            callback=callbacks,
            tb_log_name=self.agent.get_next_tensorboard_experiment_name(),
            reset_num_timesteps=False,
        )
        self.save_model()
        self.evaluate_model(validation_environment)

    def save_model(self, file_name: Optional[str] = None) -> str:
        """Save the state of the agent.

        Saves the current agent to the specified location or to a default
        location.

        Args:
            file_name (Optional[str]): Path where the agent is saved to. If
            None will see if json had a save location if also not given will
            use a default location. Defaults to None.

        Returns:
            str: Path where the agent was saved to.
        """
        if self._agent is None:
            raise RuntimeError("Please train the agent first.")
        if file_name is None:
            path = pathlib.Path(self.save_location)
        else:
            path = pathlib.Path(file_name)
        path.parent.mkdir(parents=True, exist_ok=True)
        save_path = path / "model_end.save"
        self._agent.save(save_path)
        return str(save_path)

    def evaluate_model(
        self,
        validation_env: Union[None, Environment] = None,
        throw_error_if_none: bool = False,
    ) -> None:
        """Validate the current agent.

        Evaluate the model with the parameters defined in the validation
        variable. If an agent is already loaded use this agent, else get the
        agent from the agent variable. Validation will be done inside the

        Args:
            validation_env (Union[None, Environment], optional): If validation
            environment already exists it will be used else a new validation
            environment will be created. Defaults to None.
            throw_error_if_none (bool, optional): If this is set and the
            validation variable is None an error is thrown. Defaults to False.

        Raises:
            ValidationNotSet: Thrown if validation is absolutely needed.
            If not absolutely needed the validation will not be done but no
            error will be thrown.
        """
        if self.validation:
            if validation_env is None:
                validation_env = self._create_validation_environment(True)
            if self._agent is None:
                self._agent = self.agent.get_agent(
                    validation_env.get_gym_environment()
                )
            reward, episode_length = self.validation.end_validation(
                agent=self._agent,
                environment=validation_env.get_gym_environment(),
            )
            reward_array = np.array(reward)
            mean_reward = reward_array.mean()
            reward_std = reward_array.std()
            log_str = (
                f"The end validation had the following results: "
                f"mean_reward = {mean_reward}; reward_std = {reward_std}"
            )
            self.get_logger().info(log_str)
            self.get_logger().info(f"The reward per episode was: {reward}.")
            self.get_logger().info(
                f"The length per episode was: {episode_length}"
            )
        elif throw_error_if_none:
            raise ValidationNotSet

    def _create_new_environment(self, logger_name: str):
        """Function used for multi environment training.

        Args:
            logger_name (str): name of the logger used for the rl_logger used
            for this specific environment.
        """

        def _init():
            c_env = deepcopy(self.environment)
            logging_information = {
                "logger_name": logger_name,
                "log_file_location": self.verbosity.logfile_location,
                "logging_level": self.verbosity.environment,
            }
            return c_env.get_gym_environment(
                logging_information=logging_information
            )

        return _init

    def _create_validation_environment(
        self, throw_error_if_none: bool = False
    ) -> Optional[Environment]:
        """Creates a validation environment.

        Args:
            throw_error_if_none (bool, optional):
                If this is set and the validation variable is None an error is
                thrown. Defaults to False.

        Raises:
            ValidationNotSet:
                Thrown if validation is absolutely needed. If not absolutely
                needed the validation will not be done but no error will be
                thrown.

        Returns:
            Environment:
                Validation environment. Is not a 'gym' environment but and
                ReLeSO.parser_environment.Environment. Create the gym
                environment by calling the function env.get_gym_environment()
        """
        if self.validation is None:
            # ok i have no idea why i made this thingy here but i will keep it
            # for posterity
            if throw_error_if_none:
                raise ValidationNotSet
            return None
        validation_environment = deepcopy(self.environment)

        validation_environment.set_logger_name_recursively(
            self.verbosity._environment_validation_logger
        )
        validation_environment._id = None
        validation_environment.set_validation(
            **self.validation.get_environment_validation_parameters()
        )

        return validation_environment



"""File holds definition classes for the mesh implementation."""
import pathlib
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, Literal, Optional, Union

from pydantic import Field, PrivateAttr
from pydantic.class_validators import root_validator, validator
from pydantic.types import FilePath, conint

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.util.logger import get_parser_logger
from releso.util.types import GustafMeshTypes

try:
    from gustaf.io import meshio, mixd
except ImportError as err:  # pragma: no cover
    from releso.util.module_import_raiser import ModuleImportRaiser

    meshio = ModuleImportRaiser("gustaf", err)
    mixd = ModuleImportRaiser("gustaf", err)


class MeshHierarchy(Enum):
    """Enum to define the mesh hierarchy."""

    Vertices = 0
    Edges = 1
    Faces = 2
    Volumes = 3


class MeshExporter(BaseModel):
    """Class which defines in which format and where the mesh is exported."""

    #: format to which the mesh should be exported to
    mesh_format: Literal["mixd"] = Field(default="mixd", alias="format")
    #: path to where the mesh should be exported to
    export_path: Union[str, pathlib.Path]

    #: internal variable if the export path was changed from a different value.
    #: This is the value is returned when the export path is queried.
    _export_path_changed: Optional[str] = PrivateAttr(default=None)

    @root_validator(pre=True)
    @classmethod
    def validate_path_has_correct_ending(cls, values) -> pathlib.Path:
        """Validator export_path.

        Validate, that the given path has the correct suffix and that the path
        exists.

        Args:
            values (Any): value to validate

        Raises:
            ParserException: If validation fails throws this error.

        Returns:
            pathlib.Path:
                Path to where the mesh should per default be exported to.
        """
        path = pathlib.Path(values.get("export_path")).expanduser().resolve()
        mesh_format = values.get("format")
        if mesh_format == "mixd":
            if path.suffix == "":
                path = path.with_name("_.xns")
            elif not path.suffix == ".xns":
                raise ParserException(
                    "MeshExporter",
                    "export_path",
                    "The format and the path suffix do not match. The format"
                    f" is defined as {mesh_format} but the suffix is"
                    f" {path.suffix}",
                )
        else:
            raise ParserException(
                "MeshExporter", "format", f"{mesh_format=} is not supported."
            )
        values["export_path"] = path
        return values

    def get_export_path(self) -> pathlib.Path:
        """Direct return of object variable.

        Returns:
            str:
                Path to the default export location of the mesh during
                environment operations. (So that the solver can use it.)
        """
        if not self._export_path_changed:
            self.adapt_export_path()
        return self._export_path_changed

    def adapt_export_path(self, environment_id: str = "0"):
        """If placeholder in export path insert environment_id into it.

        Args:
            environment_id (str): Environment ID
        """
        self._export_path_changed = pathlib.Path(
            str(self.export_path).format(environment_id)
        )
        self._export_path_changed.parent.mkdir(parents=True, exist_ok=True)
        self.get_logger().debug(
            f"Adapted mesh export path to the following value "
            f"{self._export_path_changed}."
        )

    def export_mesh(self, mesh: GustafMeshTypes, space_time: bool = False):
        """Exports the mesh.

        Args:
            mesh (GustafMeshTypes): _description_
            space_time (bool, optional): _description_. Defaults to False.

        Raises:
            RuntimeError: _description_
        """
        if self.mesh_format == "mixd":
            mixd.export(self._export_path_changed, mesh, space_time=space_time)
        else:
            raise RuntimeError(
                f"The requested format {self.mesh_format} is not supported."
            )


class Mesh(BaseModel):
    """Abstract class used to read in the mesh file and load it."""

    #: path to the mesh file (might be not used in case of mixd )
    path: Optional[Union[str, pathlib.Path]] = None
    #: Path to the default export location of the mesh during environment
    #: operations. (So that the solver can use it.)
    export: Optional[MeshExporter]
    #: Number of dimensions of the mesh.
    dimensions: conint(ge=1)

    @abstractmethod
    def get_mesh(self) -> GustafMeshTypes:
        """Calls the correct method to load the mesh for the gustaf library.

        Note:
            There is an error in a version of gustaf during the loading process
            . Please check with the maintainer of this package if you have
            trouble.

        Returns:
            gustaf.Mesh: Mesh in gustaf library format.
        """

    def adapt_export_path(self, environment_id: str):
        """If placeholder in export path insert environment_id into it.

        Args:
            environment_id (str): Environment ID
        """
        if self.export:
            self.export.adapt_export_path(environment_id=environment_id)

    def get_export_path(self) -> Optional[pathlib.Path]:
        """Direct return of object variable.

        Returns:
            str:
                Path to the default export location of the mesh during
                environment operations. (So that the solver can use it.)
        """
        if self.export:
            return self.export.get_export_path()

        self.get_logger().warning(
            (
                "No mesh exporter definition. Please define one if you want "
                "export functionality."
            ),
        )
        return None


class MixdMesh(Mesh):
    """Class used to read in the correct mixd mesh file and load it."""

    #: Please use either the path variable xor the mxyz variable, since if
    #: used both the used mxyz path might not be the one you think.
    mxyz_path: Optional[FilePath] = Field(
        default=None,
        description="Please use either the path variable xor the"
        " mxyz variable, since if used both the used mxyz path might not be "
        "the one you think.",
    )
    #: Please use either the path variable xor the mien variable, since if
    #: used both the used mien path might not be the one you think.
    mien_path: Optional[FilePath] = Field(
        default=None,
        description="Please use either the path variable xor the"
        " mien variable, since if used both the used mien path might not be "
        "the one you think.",
    )
    hypercube: bool = Field(
        description="If True Mesh is made of hypercubes. If False Mesh is made"
        " of simplexes (triangles).",
        default=True,
    )
    #: Number of dimensions of the mesh.
    dimensions: conint(ge=1)

    def get_mesh(self) -> GustafMeshTypes:
        """Calls the correct method to load the mesh for the gustaf library.

        Note:
            There is an error in a version of gustaf during the loading process
            . Please check with the maintainer of this package if you have
            trouble.

        Returns:
            gustaf.Mesh: Mesh in gustaf library format.
        """
        self.get_logger().debug(
            f"Loading volume mesh with mxyz file ({self.mxyz_path}) and "
            f"mien file ({self.mien_path}) ..."
        )
        mesh = mixd.load(
            simplex=not self.hypercube,
            volume=True if self.dimensions == 3 else False,
            mxyz=self.mxyz_path,
            mien=self.mien_path,
        )
        self.get_logger().info("Done loading mesh.")
        return mesh

    @root_validator
    @classmethod
    def validate_mxyz_mien_path(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validator for the whole class.

        Path to mxyz and mien files are correctly given. If mxyz_path and
        mien_path are given these are used. If not mxyz and mien path are
        generated from path.

        Args:
            values (Dict[str, Any]): All variables in a dict.

        Notes:
            Yes this function is overly complicated but so be it.

        Raises:
            ParserException: Error if the mien or mxyz files could not be
            found.

        Returns:
            Dict[str, Any]:
                All variables of the object. Hopefully with correct mien and
                mxyz paths
        """
        get_parser_logger().debug("Validating mesh file:")
        if (
            "mxyz_path" in values
            and values["mxyz_path"]
            and "mien_path" in values
            and values["mien_path"]
        ):
            get_parser_logger().info(
                "Mesh files are defined by mxyz_path and mien_path."
            )
            return values
        elif "path" in values.keys() and values["path"]:
            get_parser_logger().debug(
                "Trying to find mxyz_path and mien_path from path variable."
            )
            path = pathlib.Path(values["path"]).expanduser().resolve()
            mxyz_path = None
            mien_path = None
            if path.exists():
                if path.is_dir():
                    get_parser_logger().debug(
                        f"Given path is a directory, trying to find mxyz and "
                        f"mien files as files {path}/mxyz and {path}/mien."
                    )
                    mxyz_path = path / "mxyz.space"
                    mien_path = path / "mien"
                    if (
                        mxyz_path.exists()
                        and mxyz_path.is_file()
                        and mien_path.exists()
                        and mien_path.is_file()
                    ):
                        get_parser_logger().debug(
                            "Mesh files located with filepath given as the "
                            "root folder."
                        )
                else:  # path point to a files
                    get_parser_logger().debug(
                        "Given path is a file, trying to find mxyz and mien "
                        "files."
                    )
                    if path.suffix == ".mxyz":
                        mien_path = path.with_suffix(".mien")
                        mxyz_path = path
                        if mien_path.exists() and mien_path.is_file():
                            get_parser_logger().debug(
                                "Mesh files located with filepath given as "
                                ".mxyz file."
                            )
                        else:
                            mien_path = None
                            get_parser_logger().warning(
                                "Could not locate corresponding mien file with"
                                " path given as .mxyz file."
                            )
                    elif path.suffix == ".mien":
                        mxyz_path = path.with_suffix(".mxyz")
                        mien_path = path
                        if mxyz_path.exists() and mxyz_path.is_file():
                            get_parser_logger().debug(
                                "Mesh files located with filepath given as"
                                " .mien file."
                            )
                        else:
                            mxyz_path = None
                            get_parser_logger().warning(
                                "Could not locate corresponding mxyz file with"
                                " path given as .mien file."
                            )
                    elif path.name == "mxyz":
                        mxyz_path = path
                        mien_path = path.with_name("mien")
                        if mien_path.exists() and mien_path.is_file():
                            get_parser_logger().debug(
                                "Mesh files located with filepath given as "
                                "mxyz file."
                            )
                        else:
                            mien_path = None
                            get_parser_logger().warning(
                                "Could not locate corresponding mien file with"
                                " path given as mxyz file."
                            )
                    elif path.name == "mien":
                        mxyz_path = path.with_name("mxyz")
                        mien_path = path
                        if mxyz_path.exists() and mxyz_path.is_file():
                            get_parser_logger().debug(
                                "Mesh files located with filepath given as "
                                "mien file."
                            )
                        else:
                            mxyz_path = None
                            get_parser_logger().warning(
                                "Could not locate corresponding mxyz file with"
                                " path given as mien file."
                            )
                    else:
                        get_parser_logger().warning(
                            f"Mesh file given as existing file {path} but "
                            "could not find mxyz nor mien file."
                        )
                        raise ParserException(
                            "Mesh",
                            "path",
                            "Could not locate mien nor mxyz file path.",
                        )
                # If mien or mxyz file path are not complete throw error.
                if mxyz_path is None:
                    raise ParserException(
                        "Mesh", "path", "Could not locate mxyz file path."
                    )
                if mien_path is None:
                    raise ParserException(
                        "Mesh", "path", "Could not locate mien file path."
                    )
            elif path.suffix == ".xns":
                if (
                    path.with_suffix(".mxyz").exists()
                    and path.with_suffix(".mien").exists()
                ):
                    mien_path = path.with_suffix(".mxyz")
                    mxyz_path = path.with_suffix(".mien")

            if mien_path:
                values["mien_path"] = mien_path
                values["mxyz_path"] = mxyz_path
                values["path"] = mxyz_path.parent
            else:
                ParserException(
                    "MixdMesh", "path", "Could not locate mien and mxyz path"
                )
            return values
        raise ParserException(
            "Mesh",
            "[mien_|mxyz_|]path",
            "Could not locate the correct mien/mxyz paths.",
        )


class MeshIOMesh(Mesh):
    """Provides an interface to load meshio meshes.

    Since gustaf currently only really supports .msh files, this also only
    supports files with this ending. If and when gustaf implements support
    for additional files types already supported by meshio. This will be
    updated.
    """

    @validator("path")
    @classmethod
    def check_if_extension_is_supported_by_gustaf(cls, value) -> pathlib.Path:
        """Checks if the path given exists and the mesh type is supported.

        Args:
            value (Optional[str]): _description_

        Returns:
            pathlib.Path: Absolute path to the file.
        """
        if value:
            path = pathlib.Path(value).resolve().expanduser()
            if path.exists():
                if path.suffix != ".msh":
                    raise ParserException(
                        "MeshIOMesh", "path", "Mesh type not supported."
                    )
            else:
                raise ParserException(
                    "MeshIOMesh", "path", "Could not locate the mesh file."
                )

        else:
            raise ParserException(
                "MeshIOMesh",
                "path",
                "For MeshIO base mesh import a path must be provided.",
            )
        return path

    def get_mesh(self) -> GustafMeshTypes:
        """Calls the correct method to load the mesh for the gustaf library.

        Note:
            There is an error in a version of gustaf during the loading process
            . Please check with the maintainer of this package if you have
            trouble.

        Returns:
            gustaf.Mesh: Mesh in gustaf library format.
        """
        self.get_logger().debug(f"Loading mesh from path ({self.path}).")
        mesh = meshio.load(fname=self.path)
        final_mesh = None
        if len(mesh) > 1:
            for me in mesh:
                if final_mesh is None:
                    final_mesh = me
                    continue
                if (
                    MeshHierarchy[me.__class__.__name__].value
                    > MeshHierarchy[final_mesh.__class__.__name__].value
                ):
                    final_mesh = me
        else:
            final_mesh = mesh
        self.get_logger().info("Done loading mesh.")
        return self.delete_dimension_if_possible(final_mesh)

    def delete_dimension_if_possible(self, mesh: GustafMeshTypes):
        """Looks into the mesh and checks if the dimensionality can be reduced.

        If the bounds of one or more dimensions are equal, the dimension is
        removed from the mesh.

        Args:
            mesh (GustafMeshTypes): _description_
        """
        keep_dims = [0, 1, 2]
        if self.dimensions < mesh.vertices.shape[1]:
            bounds = mesh.bounds()
            for idx, (lower, upper) in enumerate(zip(*bounds)):
                if lower == upper:
                    self.get_logger().info(
                        f"Found empty dimension {idx}. Will remove it now."
                    )
                    keep_dims.pop(idx)
                    mesh.vertices = mesh.vertices[:, keep_dims]
        return mesh


MeshTypes = Union[MeshIOMesh, MixdMesh]



import pathlib
from typing import Literal, Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from pandas.errors import EmptyDataError
from plotly.graph_objects import Figure
from plotly.subplots import make_subplots

plotly_colors = [
    "red",
    "blue",
    "green",
    "orange",
    "purple",
    "brown",
    "pink",
    "gray",
    "olive",
    "cyan",
]
plotly_lines = ["solid", "dash", "dot", "dashdot"]


def export_figure(
    fig: Figure,
    export_path: pathlib.Path,
    default_filename: str,
) -> None:
    """Exports a given figure to a specified path.
    The function checks the file extension of the export path and saves the
    figure accordingly. If the path is a directory, it saves the figure with
    the specified default filename in that directory. If the path contains a
    file, it saves the figure with the specified extension if supported or
    prints an error message if the extension is not recognized.

    Args:
        fig (plotly.graph_objs.Figure): The figure to be exported.
        export_path (pathlib.Path): The path where the figure should be saved.
        default_filename (str): The default filename to fall back to if the
          path is a directory or if the file extension is not recognized.
    """
    # Export the plot as the suffix or as "episode_log.html"
    suffix = export_path.suffix
    if suffix == ".html":
        fig.write_html(export_path)
    elif suffix in [".png", ".jpg", ".jpeg", ".webp", ".svg", ".pdf"]:
        fig.write_image(export_path)
    elif export_path.is_dir():
        fig.write_html(export_path / default_filename)
    else:
        if not (export_path.parent / default_filename).exists():
            print(
                "File path suffix is not known nor is the file path a directory. "
                f"Saving to {export_path.parent / default_filename}"
            )
            fig.write_html(export_path.parent / default_filename)
        else:
            print(
                "File path suffix is not know nor is the file path a "
                "directory. Not saving."
            )


def plot_episode_log(
    result_folders: dict[str, pathlib.Path],
    window: int = 5,
    window_size: Union[tuple[int, int], Literal["auto"]] = "auto",
    cut_off_point: int = np.iinfo(int).max,
) -> Figure:
    """Plot one or multiple episodes to check out the training progress.

    This function can already be called once the first results are in the
    episode log file, to check out progress during training. But, in general,
    this function is normally used to compare different runs to see
    differences in the learning progress. In general all values are plotted
    over the total number of steps taken during the training. The following
    variables are shown:

    1. Episode reward - Best used to see if the training is progressing
    2. Steps per Episode - Here you can see if the number of steps to solve
       the environment decreases over time.
    3. Mean step reward - The higher the better each action actually is.
       (Calculated value episode_reward/steps_per_episode)
    4. Episode End Reason - For each run multiple lines to show why the
       episode was terminated. !ATTENTION! the window here is 100 to calculate a
       crude percentage.
    5. Seconds per Step - Time each step took to compute. Should not have much
       variation in general. But use case specific can have differences.
    6. Wall time - How much time did elapse to this time step. (Integral of
       previous plot).


    To check out the training I like to look at all the plots at once to see if
    there is anything that should concern me. For publishing most values have
    no informational gain so feel free to adapt (Remove certain sub plot, etc)
    the function.

    Author: Clemens Fricke (clemens.david.fricke@tuwien.ac.at)

    Args:
        result_folders (dict[str, pathlib.Path]): Dict of name and path to each
          experiment to visualize.
        window (int, optional): Windowing is used to smooth the plots.
          Defaults to 5.
        window_size (Union[tuple[int, int], Literal['auto']], optional): Size
         of the figure used. For html 'auto' will adapt the size to the browser
         or container size. Defaults to "auto".
        cut_off_point (int, optional): Plot the episode only to a certain
         number of time steps. Defaults to max int.

    Returns:
        fig (plotly.graph_objects.Figure): Plotly figure object with the
         requested plots for further customization or export.
    """
    end_episode = []
    df: list[pd.DataFrame] = []
    # make a separate list for just the names of the experiments
    n_env: list[str] = list(result_folders.keys())
    # set of all episode end reasons
    unique_values = set()
    for idx, folder in enumerate(result_folders.values()):
        try:
            temp_df = pd.read_csv(folder / "episode_log.csv", index_col=0)
        except EmptyDataError as err:
            print(
                f"Error occurred with data in folder {folder}. The error is {err}"
            )
            continue
        end_episode.append(
            (temp_df.loc[temp_df["total_timesteps"] < cut_off_point].index)[-1]
        )
        row = pd.to_datetime(temp_df["wall_time"])
        temp_df["wall_time"] = row
        temp_df["run_time"] = row - row.iloc[0]
        temp_df["time_delta"] = row - row.shift()
        temp_df["mean_step_reward"] = (
            temp_df["episode_reward"] / temp_df["steps_in_episode"]
        )
        unique_values.update(temp_df["episode_end_reason"].unique())
        df.append(temp_df)

    fig = make_subplots(
        rows=6, cols=1, shared_xaxes=True, vertical_spacing=0.01
    )
    # plot each experiment into each sub plot, the idx is used to determine the
    # episode end and the episode name
    for idx, dataframe in enumerate(df):
        # first subplot
        fig.add_trace(
            go.Scatter(
                x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                y=dataframe["episode_reward"]
                .iloc[: end_episode[idx]]
                .rolling(window, window)
                .mean(),
                mode="lines",
                legendgroup=f"{n_env[idx]}",  # Used so that each experiment can be disabled
                name=f"{n_env[idx]}",  # name the legend item
                line=dict(color=plotly_colors[idx]),
                showlegend=True,  # show legend only for first subplot
                customdata=dataframe.index,
                hovertemplate="(%{x:d},%{y:.2f}) Episode: %{customdata:d}",
            ),
            row=1,
            col=1,
        )
        # second subplot
        fig.add_trace(
            go.Scatter(
                x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                y=dataframe["steps_in_episode"]
                .iloc[: end_episode[idx]]
                .rolling(window, window)
                .mean(),
                mode="lines",
                legendgroup=f"{n_env[idx]}",
                name=f"{n_env[idx]}",
                line=dict(color=plotly_colors[idx]),
                showlegend=False,
            ),
            row=2,
            col=1,
        )
        # third subplot
        fig.add_trace(
            go.Scatter(
                x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                y=dataframe["mean_step_reward"]
                .iloc[: end_episode[idx]]
                .rolling(window, window)
                .mean(),
                mode="lines",
                legendgroup=f"{n_env[idx]}",
                name=f"{n_env[idx]}",
                line=dict(color=plotly_colors[idx]),
                showlegend=False,
            ),
            row=3,
            col=1,
        )
        # fourth subplot - episode end reason iterate through all reasons
        # legend for this plot is created at the end
        for marker, unique_key in zip(plotly_lines, unique_values):
            fig.add_trace(
                go.Scatter(
                    x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                    # window is 100 to calculate a crude percentage approximation
                    y=(dataframe["episode_end_reason"] == unique_key)
                    .iloc[: end_episode[idx]]
                    .rolling(100, 0)
                    .mean(),
                    line=dict(
                        color=plotly_colors[idx],
                        dash=marker,
                        width=4,
                    ),
                    legendgroup=f"{n_env[idx]}",
                    name=f"{unique_key}: {n_env[idx]}",
                    showlegend=False,
                ),
                row=4,
                col=1,
            )
        # fith subplot
        fig.add_trace(
            go.Scatter(
                x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                y=dataframe["time_delta"]
                .iloc[: end_episode[idx]]
                .dt.total_seconds()
                / dataframe["steps_in_episode"].iloc[: end_episode[idx]],
                mode="lines",
                legendgroup=f"{n_env[idx]}",
                name=f"{n_env[idx]}",
                line=dict(color=plotly_colors[idx]),
                showlegend=False,
            ),
            row=5,
            col=1,
        )
        # sixth subplot
        fig.add_trace(
            go.Scatter(
                x=dataframe["total_timesteps"].iloc[: end_episode[idx]],
                y=dataframe["run_time"]
                .iloc[: end_episode[idx]]
                .dt.total_seconds()
                / 3600.0,
                mode="lines",
                legendgroup=f"{n_env[idx]}",
                name=f"{n_env[idx]}",
                line=dict(color=plotly_colors[idx]),
                showlegend=False,
            ),
            row=6,
            col=1,
        )

    # create end reason legend
    for marker, unique_key in zip(plotly_lines, unique_values):
        fig.add_trace(
            go.Scatter(
                x=[None],
                y=[None],
                legendgroup="end_reason",
                legendgrouptitle=dict(text="Episode End Reasons"),
                name=f"{unique_key}",
                line=dict(
                    color="black",
                    dash=marker,
                    width=4,
                ),
                mode="lines",
                showlegend=True,
            ),
            row=4,
            col=1,
        )

    # adapt figure size
    fig.update_traces(line=dict(width=0.6))
    if not isinstance(window_size, list) and window_size == "auto":
        fig.update_layout(autosize=True)
    else:
        fig.update_layout(height=window_size[0], width=window_size[1])
    # update axis labels
    fig.update_xaxes(title_text="Total time steps", row=6, col=1)
    fig.update_yaxes(title_text="Episode reward", row=1, col=1)
    fig.update_yaxes(title_text="Steps<br>per episode", row=2, col=1)
    fig.update_yaxes(title_text="Mean step<br>reward", row=3, col=1)
    fig.update_yaxes(title_text="Episode<br>end reason", row=4, col=1)
    fig.update_yaxes(title_text="Seconds<br>per step [s]", row=5, col=1)
    fig.update_yaxes(title_text="Wall time [h]", row=6, col=1)

    return fig


def get_data_from_dataframe(
    df: pd.DataFrame,
    name: str,
    selector: list[tuple[str, tuple[int, int]]],
) -> pd.DataFrame:
    """Get data from a dataframe based on the provided selector.

    This function extracts specific data from a given dataframe based on the
    provided selector. The selector is a list of tuples, where each tuple
    contains a string and a tuple of two integers. The string indicates the
    type of data to extract (e.g., "obs", "info"), and the tuple of integers
    specifies the range of indices to extract. The function returns a new
    dataframe containing the extracted data.
    The function is designed to handle different types of data extraction,
    including observations and information from the dataframe. It uses
    recursion to navigate through the dataframe structure (and subsequent dict)
    and extract the desired data.

    Args:
        df (pd.DataFrame): StepDataFrame containing the data to be extracted.
        name (str): Name of field the extracted data should be stored in.
        selector (list[tuple[str, tuple[int, int]]]): List of tuples
          containing the type of data to extract and the range of indices.

    Returns:
        pd.DataFrame: DataFrame containing the extracted data.
    """

    def get_elem_from_crums(df: pd.DataFrame, crums: list[str]) -> np.ndarray:
        """Get the element from the dataframe/dictionary based on the crums."""
        if len(crums) == 1:
            return df.apply(lambda x: x[crums[0]])
        else:
            return get_elem_from_crums(
                df.apply(lambda x: x[crums[0]]), crums[1:]
            )

    objective_id = 0
    ret_df = []
    for element in selector:
        if element[0] == "obs":
            selected = np.vstack(df["obs"].values).T[
                element[1][0] : element[1][1]
            ]
            ret_df.append(
                pd.DataFrame(
                    selected.T,
                    columns=[
                        f"{name}_{objective_id + i}"
                        for i in range(len(selected))
                    ],
                )
            )
            objective_id += len(selected)
        elif element[0].startswith("infos"):
            if element[0].endswith("geometry_information"):
                selected = np.vstack(
                    df["infos"].apply(
                        lambda x: np.array(x["geometry_information"]).flatten()
                    )
                ).T[element[1][0] : element[1][1]]
            else:
                crums = element[0].split(".")
                selected = np.vstack(
                    get_elem_from_crums(df["infos"], crums[1:])
                ).T[element[1][0] : element[1][1]]
            ret_df.append(
                pd.DataFrame(
                    selected.T,
                    columns=[
                        f"{name}_{objective_id + i}"
                        for i in range(len(selected))
                    ],
                )
            )
            objective_id += len(selected)
    return pd.concat(ret_df, axis=1)


def plot_step_log(
    step_log_file: pathlib.Path,
    env_id: int,
    episode_start: int = 0,
    episode_end: int = np.iinfo(int).max,
    episode_step: int = 1,
    figure_size: Union[tuple[int, int], Literal["auto"]] = "auto",
    objective_observation: list[tuple[str, tuple[int, int]]] = [
        ("obs", (0, 1))
    ],
    design_variable: list[tuple[str, tuple[int, int]]] = [
        ("obs", (1, None)),
    ],
) -> Figure:
    """Plot the step log data of a single run for multiple episodes.

    This function is used to visualize the step log data of a single run
    with the intention of understanding the training progression and the
    policy that is learned by the agent.
    It creates an interactive plot with two subplots:

    1. The first subplot shows the reward and objective value(s) over the number
       of timesteps within the current episode.
    2. The second subplot shows the values of the design variables (observations)
       over the number of timesteps within the current episode.
       The plot is interactive, allowing users to select different episodes via a
       slider to view the corresponding data.

    Since plotly does not recompute data when the user interacts with the
    plot, the data for all episodes is precomputed and stored in the figure.
    This can result in large file sizes, especially when the optimization
    contained a lot of design parameters and when the chosen episode range is
    large. Please be aware of this when using the function.

    Author: Daniel Wolff (d.wolff@unibw.de)

    Args:
        step_log_file (pathlib.Path): Path to the step log file.
        env_id (int): ID of the environment whose results should be visualized.
          This parameter is only relevant in the case of multi-environment
          training. If the training did not use multiple environments, this
          parameter should be set to 0.
        episode_start (int, optional): First episode that should be included in
          the visualization. Defaults to 0.
        episode_end (int, optional): Last episode that should be included in
          the visualization. If None, the maximum episode number is used.
          Defaults to None.
        episode_step (int, optional): Step size for selecting episodes.
          Defaults to 1, which means that every episode between the starting
          and final episode are visualized.
        figure_size (Union[tuple[int, int], Literal['auto']], optional):
          Size of the figure. If 'auto', the size will be adjusted to fit the
          container. Defaults to 'auto'.
        objective_observation (list[tuple[str, tuple[int, int]]]): List of
          tuples containing information from which column of the steplog data
          to extract the information about the objective (each row in a column
          generally contains a list of values) and the indices which entries
          from that list are supposed to be to extracted.
        design_variable (list[tuple[str, tuple[int, int]]]): List of tuples
          containing information from which column of the steplog data to
          extract the information about the design variables (each row in a
          column generally contains a list of values) and the indices which
          entries from that list are supposed to be to extracted.
    Returns:
        plotly.graph_objects.Figure: A Plotly figure object containing the
          interactive plot for further customization or export.
    """
    # Load the step log data from the provided path
    try:
        df_raw = pd.read_json(step_log_file, lines=True)
    except RuntimeError as err:
        print(
            f"Error occurred when trying to read {step_log_file}. The error is {err}"
        )
        return

    # Process the data for the visualization

    # Extract scalar reward and observation (we use new_obs for visualization)
    # The underlying assumption of the extraction is that we are always dealing
    # with a vectorized environment, i.e., that the rewards and observations
    # corresponding to each environment are collected in a list
    df_raw["episodes"] = df_raw["episodes"].apply(lambda x: x[env_id])
    df_raw["reward"] = df_raw["rewards"].apply(lambda x: x[env_id])
    df_raw["obs"] = df_raw["new_obs"].apply(lambda x: x[env_id])
    has_info = "infos" in df_raw.columns
    if has_info:
        df_raw["infos"] = df_raw["infos"].apply(lambda x: x[env_id])

    # Convert obs vector into columns
    # obs_array = np.vstack(df_raw["obs"].values)
    # obs_df = pd.DataFrame(
    #     obs_array, columns=[f"obs_{i}" for i in range(obs_array.shape[1])]
    # )
    # Extract the objective observation
    objectives = get_data_from_dataframe(
        df_raw,
        "objective",
        objective_observation,
    )

    # Extract the design variables
    design_vars = get_data_from_dataframe(
        df_raw,
        "design_variable",
        design_variable,
    )

    # Combine everything
    df = pd.concat(
        [df_raw["episodes"], df_raw["reward"], objectives, design_vars], axis=1
    )

    # Filter only the selected episodes
    max_idx = df["episodes"].max()
    if episode_end is None or episode_end > max_idx:
        episode_end = max_idx
    # selected episode numbers do not necessarily match the filter applied. The
    # filter does not directly filter for episode number but filters the whole
    # list of episodes, which can have missing episodes, due to episodes
    # generated outside the environment id chosen.
    selected_episodes = df["episodes"].unique()[
        episode_start : (episode_end + 1) : episode_step
    ]
    df = df[df["episodes"].isin(selected_episodes)]

    # Create the interactive visualization

    # Choose which obs dimensions to show in bottom subplot
    design_var_names = df.columns[
        df.columns.str.contains("design_variable_")
    ].tolist()

    # Choose which obs dimensions to show in top subplot
    objective_names = df.columns[
        df.columns.str.contains("objective_")
    ].tolist()

    # Get all unique episodes
    episodes = df["episodes"].unique()

    # Create subplot layout
    fig = make_subplots(
        rows=2,
        cols=1,
        shared_xaxes=True,
        specs=[
            [{"secondary_y": True}],
            [{}],
        ],  # Enable secondary y-axis for row 1
        subplot_titles=("Reward and Objective", "Design variables"),
        vertical_spacing=0.1,
    )

    # Build one trace group per episode
    for ep in episodes:
        ep_data = df[df["episodes"] == ep]
        steps_per_episode = list(range(len(ep_data)))

        # First subplot: reward and objective
        for i, obs_dim in enumerate(objective_names):
            fig.add_trace(
                go.Scatter(
                    x=steps_per_episode,
                    y=ep_data[obs_dim],
                    name=f"Objective_{i} (Ep. {ep})",
                    visible=(ep == episode_start),
                    # line=dict(color="green"),
                ),
                row=1,
                col=1,
            )
        fig.add_trace(
            go.Scatter(
                x=steps_per_episode,
                y=ep_data["reward"],
                name=f"Reward (Ep. {ep})",
                visible=(ep == episode_start),
                line=dict(color="blue"),
            ),
            row=1,
            col=1,
            secondary_y=True,
        )

        # Second subplot: selected observation dimensions
        for j, dim in enumerate(design_var_names):
            fig.add_trace(
                go.Scatter(
                    x=steps_per_episode,
                    y=ep_data[dim],
                    name=f"Design Var. {j} (Ep. {ep})",
                    visible=(ep == episode_start),
                ),  # one trace per dimension
                row=2,
                col=1,
            )

    # once all traces are added, set up slider steps
    sliders = []
    for id, ep in enumerate(episodes):
        # Create slider step
        sliders.append({
            "label": f"Episode {ep}",
            "method": "update",
            "args": [
                {"visible": [False] * len(fig.data)},
                {"title": f"Steplog Evaluation — Episode {ep}"},
            ],
        })
        # Make the traces for this episode visible
        n_traces_per_episode = 1 + len(objective_names) + len(design_var_names)
        for k in range(n_traces_per_episode):
            sliders[id]["args"][0]["visible"][
                id * n_traces_per_episode + k
            ] = True

    # Set up slider control
    fig.update_layout(
        sliders=[{"active": 0, "pad": {"t": 50}, "steps": sliders}],
        title="Steplog Evaluation",
        showlegend=True,
    )

    # rescale the figure
    if not isinstance(figure_size, list) and figure_size == "auto":
        fig.update_layout(autosize=True)
    else:
        fig.update_layout(height=figure_size[0], width=figure_size[1])

    # Set y-axis labbels
    fig.update_yaxes(title_text="Objective", row=1, col=1, secondary_y=False)
    fig.update_yaxes(title_text="Reward", row=1, col=1, secondary_y=True)
    fig.update_yaxes(title_text="Observation Value", row=2, col=1)

    return fig



import numpy as np
import pytest
from gymnasium import spaces
from pydantic import ValidationError

from releso.geometry import FFDGeometry, Geometry
from releso.shape_parameterization import ShapeDefinition
from releso.spline import BSplineDefinition, NURBSDefinition
from releso.util.logger import VerbosityLevel


@pytest.mark.parametrize(
    (
        "shape_definition",
        "action_based",
        "discrete_action",
        "reset_random",
        "correct_type",
        "n_action_variables",
    ),
    [
        (
            "default_shape",
            None,
            None,
            None,
            ShapeDefinition,
            2,
        ),  # test that default is correct
        ("default_shape", True, True, True, ShapeDefinition, 2),
        ("default_shape", False, False, False, ShapeDefinition, 2),
        ("bspline_shape", None, False, None, BSplineDefinition, 18),
        ("nurbs_shape", None, False, None, NURBSDefinition, 20),
    ],
)
def test_geometry_init(
    shape_definition,
    action_based,
    discrete_action,
    reset_random,
    correct_type,
    n_action_variables,
    dir_save_location,
    caplog,
    request,
):
    call_dict = {
        "shape_definition": request.getfixturevalue(shape_definition),
        "save_location": dir_save_location,
    }
    if action_based is not None:
        call_dict["action_based_observation"] = action_based
    if discrete_action is not None:
        call_dict["discrete_actions"] = discrete_action
    if reset_random is not None:
        call_dict["reset_with_random_action_values"] = reset_random
    geometry = Geometry(**call_dict)
    assert isinstance(geometry.shape_definition, correct_type)
    assert geometry.action_based_observation == (
        action_based if action_based is not None else True
    )
    assert geometry.discrete_actions == (
        discrete_action if discrete_action is not None else True
    )
    assert geometry.reset_with_random_action_values == (
        reset_random if reset_random is not None else False
    )

    # check get all actions
    geometry.setup("id")

    # control points
    assert (
        geometry.get_parameter_values()
        == geometry.shape_definition.get_parameter_values()
    )

    assert geometry.is_geometry_changed() is False
    assert geometry.apply() == geometry.get_parameter_values()

    original_cps = geometry.get_parameter_values()

    # actions
    assert len(geometry._actions) == len(
        geometry.shape_definition.get_actions()
    )
    assert len(geometry._actions) == n_action_variables

    act_def = geometry.get_action_definition()
    if geometry.discrete_actions:
        assert isinstance(act_def, spaces.Discrete)
        assert act_def.n == 2 * len(geometry._actions)
        geometry.apply_action(3)
    else:
        assert isinstance(act_def, spaces.Box)
        assert act_def.shape == (len(geometry._actions),)
        geometry.apply_action(np.random.rand(len(geometry._actions)))
    assert geometry.is_geometry_changed() is True

    # observation
    if geometry.action_based_observation:
        assert geometry.get_observation_definition()[1].shape == (
            len(geometry._actions),
        )
        assert len(geometry.get_observation()) == len(geometry._actions)
    else:
        with caplog.at_level(VerbosityLevel.WARNING):
            geometry.get_observation_definition()
            assert (
                "Observation space is accessed which should not happen."
                in caplog.text
            )
            assert geometry.get_observation() is None

    # reset
    current_cps = geometry.get_parameter_values()
    assert current_cps != original_cps
    if geometry.reset_with_random_action_values:
        reset_cps = geometry.reset()
        assert reset_cps != current_cps
        assert reset_cps != original_cps
    else:
        assert geometry.reset() == original_cps

    # random action with seed produces same result
    set_1 = geometry.apply_random_action("asd")
    set_2 = geometry.apply_random_action("asd")
    assert np.allclose(set_1, set_2)


@pytest.mark.parametrize(
    (
        "shape_definition",
        "action_based",
        "discrete_action",
        "reset_random",
        "load_sample_file",
        "export_mesh",
        "error",
    ),
    [
        (
            "default_shape",
            False,
            False,
            False,
            "volumes/tet/3DBrickTet.msh",
            None,
            "FFD can only be performed with a splinepy Spline",
        ),
        (
            "bspline_shape",
            None,
            False,
            None,
            "faces/quad/2DChannelQuad.msh",
            {"format": "mixd", "export_path": "test"},
            False,
        ),
        (
            "nurbs_shape",
            None,
            False,
            None,
            "faces/quad/2DChannelQuad.msh",
            None,
            False,
        ),
    ],
    indirect=["load_sample_file"],
)
def test_ffd_geometry_init(
    shape_definition,
    action_based,
    discrete_action,
    reset_random,
    load_sample_file,
    export_mesh,
    error,
    dir_save_location,
    caplog,
    request,
):
    call_dict = {
        "shape_definition": request.getfixturevalue(shape_definition),
        "save_location": dir_save_location,
        "mesh": {
            "path": load_sample_file,
            "save_location": dir_save_location,
            "dimensions": 2,
        },
    }
    if export_mesh is not None:
        export_mesh["save_location"] = dir_save_location
        call_dict["export_mesh"] = export_mesh
    if action_based is not None:
        call_dict["action_based_observation"] = action_based
    if discrete_action is not None:
        call_dict["discrete_actions"] = discrete_action
    if reset_random is not None:
        call_dict["reset_with_random_action_values"] = reset_random

    if error:
        with pytest.raises(ValidationError) as err:
            geometry = Geometry(**call_dict)
            assert error in str(err.value)
        return

    geometry = FFDGeometry(**call_dict)
    assert geometry.action_based_observation == (
        action_based if action_based is not None else True
    )
    assert geometry.discrete_actions == (
        discrete_action if discrete_action is not None else True
    )
    assert geometry.reset_with_random_action_values == (
        reset_random if reset_random is not None else False
    )

    with caplog.at_level(VerbosityLevel.INFO):
        # This warning should happen
        geometry.setup("id")
        assert "Found empty dimension" in caplog.text
    # control points
    assert (
        geometry.get_parameter_values()
        == geometry.shape_definition.get_parameter_values()
    )

    assert geometry.is_geometry_changed() is False
    geometry.apply()

    _ = geometry.get_parameter_values()

    # actions
    assert len(geometry._actions) == len(
        geometry.shape_definition.get_actions()
    )
    act_def = geometry.get_action_definition()
    if geometry.discrete_actions:
        assert isinstance(act_def, spaces.Discrete)
        assert act_def.n == 2 * len(geometry._actions)
        geometry.apply_action(3)
    else:
        assert isinstance(act_def, spaces.Box)
        assert act_def.shape == (len(geometry._actions),)
        geometry.apply_action(np.random.rand(len(geometry._actions)))
    assert geometry.is_geometry_changed()


@pytest.mark.parametrize(
    (
        "shape_definition",
        "correct_type",
        "n_action_variables",
        "observation_shape",
    ),
    [
        (
            "default_shape",
            ShapeDefinition,
            2,
            [5, 5],
        ),  # test that default is correct
        (
            ["default_shape", "default_shape"],
            [ShapeDefinition, ShapeDefinition],
            4,
            [[5, 5], [5, 5]],
        ),
        (
            ["default_shape", "bspline_shape"],
            [ShapeDefinition, BSplineDefinition],
            20,
            [[5, 5], [2, 2, 2, 2, 2, 2, 2, 2, 2]],
        ),
        (
            ["default_shape", "nurbs_shape"],
            [ShapeDefinition, NURBSDefinition],
            22,
            [[5, 5], [2, 2, 2, 2, 2, 2, 2, 2, 2, 9]],
        ),
        # if nurbs and b splines are mixed the shape is always a nurbs.
        (
            ["bspline_shape", "nurbs_shape", "default_shape"],
            [BSplineDefinition, NURBSDefinition, ShapeDefinition],
            40,
            [
                [2, 2, 2, 2, 2, 2, 2, 2, 2],
                [2, 2, 2, 2, 2, 2, 2, 2, 2, 9],
                [5, 5],
            ],
        ),
    ],
)
def test_geometry_shape_list(
    shape_definition,
    correct_type,
    n_action_variables,
    observation_shape,
    dir_save_location,
    request,
):
    call_dict = {
        "save_location": dir_save_location,
    }
    if isinstance(shape_definition, list):
        call_dict["shape_definition"] = [
            request.getfixturevalue(shape) for shape in shape_definition
        ]
    else:
        call_dict["shape_definition"] = request.getfixturevalue(
            shape_definition
        )

    geometry = Geometry(**call_dict)
    # correct type of shape definition
    if isinstance(correct_type, list):
        assert isinstance(geometry.shape_definition, list), (
            "Shape definition should be a list."
        )
        assert len(geometry.shape_definition) == len(correct_type), (
            "Shape definition should have the correct length."
        )
        for shape, correct in zip(geometry.shape_definition, correct_type):
            assert isinstance(shape, correct), (
                f"Shape definition should be of type {correct}, but is {[type(s) for s in geometry.shape_definition]} {correct_type}."
            )
    else:
        assert isinstance(geometry.shape_definition, correct_type), (
            f"Shape definition should be of type {correct_type}, but is {type(geometry.shape_definition)}."
        )

    # correct action shape

    geometry.setup("id")

    assert len(geometry._actions) == n_action_variables

    # correct observation shape
    def check_shape(item, shape):
        assert len(item) == len(shape), (
            f"Length of item and b must be equal. Got {item} and {shape}."
        )
        for i, s in zip(item, shape):
            if isinstance(s, list):
                check_shape(i, s)

    param_values = geometry.get_parameter_values()

    check_shape(param_values, observation_shape)



import pytest

from releso.util.module_import_raiser import ModuleImportRaiser


def test_module_import_raiser_additional_error_message():
    raiser = ModuleImportRaiser("module_name", "additional error message")
    with pytest.raises(ImportError) as excinfo:
        raiser()
    assert "additional error message" in str(excinfo.value)


def test_module_import_raiser():
    raiser = ModuleImportRaiser("module_name")
    # test call functionality
    with pytest.raises(ImportError) as excinfo:
        raiser()
    assert "module_name" in str(excinfo.value)
    # changed call
    with pytest.raises(ImportError):
        raiser("additional error message")
    # getattr
    with pytest.raises(ImportError):
        raiser.test  # noqa: B018
    # setattr
    with pytest.raises(ImportError):
        raiser.test = "hallo"
    # setattr
    with pytest.raises(ImportError):
        raiser["hallo"] = "hallo"
    with pytest.raises(ImportError):
        raiser["hallo"]



"""Module Raiser optional dependency for modules."""
from typing import Any, Optional


class ModuleImportRaiser:
    """Import error deferrer until it is actually called.

    Class used to have better import error handling in the case that a package
    package is not installed. This is necessary due to that some packages are
    not a dependency of `splinepy`, but some parts require them to function.
    Examples are `splinepy`, `torchvision`, and `imageio`.
    """

    def __init__(
        self, lib_name: str, error_mesg: Optional[str] = None
    ) -> None:
        """Constructor of object of class ModuleImportRaiser.

        Args:
            lib_name (str): Name of the library which can not be loaded. Will
                be inserted into the error message of the deferred import
                Error. Is not checked for correctness.
            error_mesg (Optional[str], optional): Original error msg. Defaults
                to None.
        """
        self._message = str(
            "Parts of the requested functionality in ReLeSO depend on the "
            f"external `{lib_name}` package which could not be found on "
            "your system. Please refer to the installation instructions "
            "for more information."
            f"{f'Original error message {error_mesg}' if error_mesg else ''}"
        )

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Dummy method for object(args, kwargs).

        Is called when the object is called by object(). Will notify the user,
        that the functionality is not accessible and how to proceed to access
        the functionality.
        """
        raise ImportError(self._message)

    def __getattr__(self, __name: str) -> Any:
        """Dummy method for object.__name.

        Is called when any attribute of the object is accessed by object.attr.
        Will notify the user, that the functionality is not accessible and how
        to proceed to access the functionality.
        """
        # if __name == "_ModuleImportRaiser__message":
        #     return object.__getattr__(self, __name[-8:])
        # else:
        raise ImportError(self._message)

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Dummy method for object.__name = __value.

        Is called when any attribute of the object is set by object.attr = new.
        Will notify the user, that the functionality is not accessible and how
        to proceed to access the functionality.
        """
        if __name == "_message":
            object.__setattr__(self, __name, __value)
        else:
            raise ImportError(self._message)

    def __getitem__(self, key):
        """Dummy method for object[key].

        Is called when the object is subscripted object[x]. Will notify the
        user, that the functionality is not accessible and how to proceed to
        access the functionality.
        """
        raise ImportError(self._message)

    def __setitem__(self, key, value):
        """Dummy method for object[key] = value.

        Is called when the object is subscripted object[x]. Will notify the
        user, that the functionality is not accessible and how to proceed to
        access the functionality.
        """
        raise ImportError(self._message)
